"""
Local free energy of lnPi (:mod:`~lnpy.lnpienergy`)
===================================================
"""

from __future__ import annotations

import itertools
import warnings
from typing import TYPE_CHECKING, cast, overload

import numpy as np
import pandas as pd
import xarray as xr
from module_utilities import cached

from .docstrings import docfiller
from .utils import get_tqdm_calc as get_tqdm
from .utils import labels_to_masks, masks_change_convention, parallel_map_func_starargs

if TYPE_CHECKING:
    from typing import Any, Iterable, Literal, Sequence, Union

    from ._typing import MaskConvention, MyNDArray
    from ._typing_compat import Self
    from .lnpiseries import lnPiCollection

    _FindBoundariesMode = Literal["thick", "inner", "outer", "subpixel"]
    _FindBoundariesMethod = Literal["exact", "approx"]
    _Extrema = Literal["min", "max"]

    _FillArg = Union[int, None]
    _FillVal = Union[_FillArg, float]
    _ExtremaArg = Union[int, "tuple[int, ...]", None]


@docfiller.decorate
def find_boundaries(
    masks: Sequence[MyNDArray],
    mode: _FindBoundariesMode = "thick",
    connectivity: int | None = None,
    **kws: Any,
) -> list[MyNDArray]:
    """
    Find boundary region for masks

    Parameters
    ----------
    {masks_image}
    mode : str, default="thick"
        mode to use in :func:`skimage.segmentation.find_boundaries`
    {find_boundary_connectivity}


    Returns
    -------
    boundaries : list of array of bool
        Boundaries of each mask.

    See Also
    --------
    skimage.segmentation.find_boundaries

    """
    from skimage import segmentation

    if connectivity is None:
        connectivity = np.asarray(masks[0]).ndim

    return [
        segmentation.find_boundaries(m, connectivity=connectivity, mode=mode, **kws)
        for m in masks
    ]


@overload
def find_boundaries_overlap(
    masks: Sequence[MyNDArray],
    *,
    boundaries: list[MyNDArray] | None = ...,
    flag_none: bool = ...,
    mode: _FindBoundariesMode = ...,
    connectivity: int | None = ...,
    method: Literal["exact"] = ...,
) -> dict[tuple[int, int, int], MyNDArray | None]: ...


@overload
def find_boundaries_overlap(
    masks: Sequence[MyNDArray],
    *,
    boundaries: list[MyNDArray] | None = ...,
    flag_none: bool = ...,
    mode: _FindBoundariesMode = ...,
    connectivity: int | None = ...,
    method: Literal["approx"],
) -> dict[tuple[int, int], MyNDArray | None]: ...


# @overload
# def find_boundaries_overlap(
#     masks: list[MyNDArray],
#     *,
#     boundaries: None=...,
#     flag_none: bool=...,
#     mode: str=...,
#     connectivity: int | None = ...,
#     method: str = ...,
# ) -> dict[tuple[int, int] | tuple[int, int, int], MyNDArray | None ]: ...


@docfiller.decorate
def find_boundaries_overlap(
    masks: Sequence[MyNDArray],
    *,
    boundaries: list[MyNDArray] | None = None,
    flag_none: bool = True,
    mode: _FindBoundariesMode = "thick",
    connectivity: int | None = None,
    method: _FindBoundariesMethod = "exact",
) -> (
    dict[tuple[int, int, int], MyNDArray | None]
    | dict[tuple[int, int], MyNDArray | None]
):
    """
    Find regions where boundaries overlap

    Parameters
    ----------
    {masks_image}
    boundaries : list of array of bool, optional
        If supplied, use these areas as boundaries.  Otherwise, calculate
        boundaries using `find_boundaries`
    flag_none : bool, default=True
        if True, replace overlap with None if no overlap between regions
    {find_boundary_connectivity}
    method : str, {{'approx', 'exact'}}
        * approx : consider regions where boundaries overlap and in one of the two region
        * exact : consider regions where boundaries overlap

    Returns
    -------
    overlap : dict of array of bool
        overlap[i, j] = region of overlap between boundaries and masks in regions i and j


    See Also
    --------
    find_boundaries
    """

    n = len(masks)
    possible_methods = {"approx", "exact"}

    if method not in possible_methods:
        msg = f"{method=} not in {possible_methods}"
        raise ValueError(msg)

    if boundaries is None:
        boundaries = find_boundaries(masks, mode=mode, connectivity=connectivity)

    # assert n == len(boundaries)
    if n != len(boundaries):
        msg = f"{boundaries=} must have length {n}."
        raise ValueError(msg)

    def _get_approx() -> dict[tuple[int, int], MyNDArray | None]:
        result: dict[tuple[int, int], MyNDArray | None] = {}
        for i, j in itertools.combinations(range(n), 2):
            # overlap region is where boundaries overlap, and in one of the masks
            overlap = (boundaries[i] & boundaries[j]) & (masks[i] | masks[j])
            if flag_none and not np.any(overlap):
                result[i, j] = None
            else:
                result[i, j] = overlap
        return result

    def _get_exact() -> dict[tuple[int, int, int], MyNDArray | None]:
        result: dict[tuple[int, int, int], MyNDArray | None] = {}
        for i, j in itertools.combinations(range(n), 2):
            boundaries_overlap = boundaries[i] & boundaries[j]
            for index, m in enumerate([masks[i], masks[j]]):
                overlap = boundaries_overlap & m
                if flag_none and not np.any(overlap):
                    result[i, j, index] = None
                else:
                    result[i, j, index] = overlap
        return result

    if method == "approx":
        return _get_approx()
    if method == "exact":
        return _get_exact()

    msg = f"unknown method={method}"
    raise ValueError(msg)


@docfiller.decorate
def find_masked_extrema(
    data: MyNDArray,
    masks: Sequence[MyNDArray | None],
    convention: MaskConvention = "image",
    extrema: _Extrema = "max",
    fill_val: _FillVal = np.nan,
    fill_arg: _FillArg = None,
    unravel: bool = True,
) -> tuple[list[_ExtremaArg], MyNDArray]:
    """
    Find position and value of extrema of masked data.

    Parameters
    ----------
    data : ndarray
        input data to consider
    {masks_general}
    {mask_convention}
    extrema : {{'max','min}}
        Type of extrema to calculate
    fill_val : scalar, default=nan
        Value to fill for `out_arg` for empty `mask`
    fill_arg : scalar, optional
        Value to fill for `out_val` for empty `mask`
    unravel : bool, default=True
        If True, unravel flat index to multi dimensional index.


    Returns
    -------
    out_arg : list of int
        index of extrema, one for each `mask`
    out_val : ndarray
        value of extrema, one for each `mask`
    """

    fill_val = fill_val or np.nan

    if extrema == "max":
        func = np.argmax
    elif extrema == "min":
        func = np.argmin
    else:
        msg = 'extrema must be on of {"min", "max}'
        raise ValueError(msg)

    masks = masks_change_convention(masks, convention, "image")

    data_flat = data.reshape(-1)
    positions_flat = np.arange(data.size)

    out_val = []
    out_arg = []

    arg: _ExtremaArg
    val: _FillVal

    for mask in masks:
        if mask is None or not np.any(mask):
            arg = fill_arg
            val = fill_val
        else:
            mask_flat = mask.reshape(-1)
            arg = positions_flat[mask_flat][func(data_flat[mask_flat])]
            val = data_flat[arg]  # type: ignore[assignment]

            if unravel:
                arg = np.unravel_index(arg, data.shape)  # type: ignore[assignment,arg-type]

        out_arg.append(arg)
        out_val.append(val)

    return out_arg, np.array(out_val)


@docfiller.decorate
def merge_regions(
    w_tran: MyNDArray,
    w_min: MyNDArray,
    masks: Sequence[MyNDArray],
    nfeature_max: int | None = None,
    efac: float = 1.0,
    force: bool = True,
    convention: MaskConvention = "image",
    warn: bool = True,
) -> tuple[Sequence[MyNDArray], MyNDArray, MyNDArray]:
    r"""
    Merge labels where free energy energy barrier < efac.

    The scaled free energy :math:`w = \beta f = - \ln \Pi ` is analyzed.
    Anywhere where :math:`\Delta w = \text{{w_tran}} - \text{{w_min}} < \text{{efac}}` will be merged into a single phase.  Here,
    ``w_trans`` is the transition energy between phases (i.e., the minimum value of `lnPi`` between regions`) and ``w_min`` is the minimum energy (the maximum `lnPi`) for a phase.



    Parameters
    ----------
    w_tran : array
        Shape of array is ``(n, n)``, where ``n`` is the number of unique regions/phases.
    This is the transitional free energy ()
    w_min : array
        Shape of array is ``(n, 1)``.
    {masks_general}
    nfeature_max : int
        maximum number of features/phases to allow
    efac : float, default=0.5
        Energy difference to merge on. When ``w_trans[i, j] - w_min[i] < efac``, phases
        ``i`` and ``j`` will be merged together.
    force : bool, default=True
        if True, then keep going until nfeature <= nfeature_max
        even if min_val > efac.
    {mask_convention}
    warn : bool, default=True
        if True, give warning messages

    Returns
    -------
    masks : list of ndarray of bool
        output masks using `convention`
    w_trans : array
        transition energy for new masks
    w_min : array
        free energy minima of new masks
    """

    nfeature = len(masks)
    if nfeature_max is None:
        nfeature_max = nfeature

    w_tran = np.array(w_tran, copy=True)
    w_min = np.array(w_min, copy=True)

    # keep track of keep/kill
    mapping = dict(enumerate(masks))
    for _cnt in range(nfeature):
        # number of finite minima
        nfeature = len(mapping)

        de = w_tran - w_min

        min_arg = np.unravel_index(np.nanargmin(de), de.shape)
        min_val = de[min_arg]

        if min_val > efac:
            if not force:
                if warn and nfeature > nfeature_max:
                    warnings.warn(
                        "min_val > efac, but still too many phases",
                        Warning,
                        stacklevel=2,
                    )
                break
            if nfeature <= nfeature_max:
                break

        idx_keep, idx_kill = min_arg
        # keep the one with lower energy
        if w_min[idx_keep, 0] > w_min[idx_kill, 0]:
            idx_keep, idx_kill = idx_kill, idx_keep

        # transition from idx_keep to any other phase equals the minimum transition
        # from either idx_keep or idx_kill to that other phase
        new_tran = w_tran[[idx_keep, idx_kill], :].min(axis=0)
        new_tran[idx_keep] = np.inf
        w_tran[idx_keep, :] = w_tran[:, idx_keep] = new_tran

        # get rid of old one
        w_tran[idx_kill, :] = w_tran[:, idx_kill] = np.inf

        # new mask
        mapping[idx_keep] |= mapping[idx_kill]  # type: ignore[index]
        del mapping[idx_kill]  # type: ignore[arg-type]

    # from mapping create some new stuff
    # new w/de
    idx_min = list(mapping.keys())
    w_min = w_min[idx_min]

    idx_tran = np.ix_(*(idx_min,) * 2)
    w_tran = w_tran[idx_tran]

    # get masks
    masks = [mapping[i] for i in idx_min]

    # optionally convert image
    masks = masks_change_convention(masks, True, convention)

    return masks, w_tran, w_min


@docfiller.decorate
class wFreeEnergy:  # noqa: N801
    r"""
    Analysis of local free energy :math:`w = \beta f = - \ln \Pi`.

    Parameters
    ----------
    data : ndarray
        lnPi data
    {masks_general}
    {mask_convention}
    {find_boundary_connectivity}
    index : sequence of int, optional
        Optional index to apply to phases.
        Not yet fully supported.

    Notes
    -----
    this class uses the 'image' convention for masks.  That is
    where `mask == True` indicates locations in the phase/region,
    and `mask == False` indicates locations not in the phase/region.
    This is opposite the convention used in :mod:`numpy.ma`.
    """

    def __init__(
        self,
        data: MyNDArray,
        masks: Sequence[MyNDArray],
        convention: MaskConvention = "image",
        connectivity: int | None = None,
        index: Sequence[int] | MyNDArray | None = None,
    ) -> None:
        self.data = np.asarray(data)

        # make sure masks in image convention
        self.masks = masks_change_convention(masks, convention, "image")

        if index is None:
            self.index = np.arange(self.nfeature)
        else:
            self.index = np.asarray(self.index, dtype=np.int_)

        if connectivity is None:
            connectivity = self.data.ndim
        self.connectivity = connectivity

        self._cache: dict[str, Any] = {}

    @property
    def nfeature(self) -> int:
        """Number of features/regions/phases"""
        return len(self.masks)

    @classmethod
    @docfiller.decorate
    def from_labels(
        cls,
        data: MyNDArray,
        labels: MyNDArray,
        connectivity: int | None = None,
        features: Sequence[int] | None = None,
        include_boundary: bool = False,
        **kwargs: Any,
    ) -> Self:
        """
        Create wFreeEnergy from labels

        Parameters
        ----------
        data : ndarray
            lnPi data
        {labels}
        {find_boundary_connectivity}
        {features}
        {include_boundary}
        **kwargs
            Extra arguments to :func:`~lnpy.utils.labels_to_masks`

        Returns
        -------
        out : :class:`wFreeEnergy`

        See Also
        --------
        lnpy.utils.labels_to_masks
        """
        masks, features = labels_to_masks(
            labels,
            features=features,
            convention="image",
            include_boundary=include_boundary,
            **kwargs,
        )
        return cls(data=data, masks=masks, connectivity=connectivity)

    @cached.prop
    def _data_max(self) -> tuple[list[_ExtremaArg], MyNDArray]:
        """For lnPi data, find absolute argmax and max"""
        return find_masked_extrema(self.data, self.masks)

    @cached.meth
    def _boundary_max(
        self, method: Literal["approx", "exact"] = "exact", fill_value: float = np.nan
    ) -> tuple[dict[tuple[int, int], _ExtremaArg], MyNDArray]:
        """
        Find argmax along boundaries of regions.
        Corresponds to argmin(w)

        if method == 'exact', then find the boundary of each region
        and find max.  then find min of those maxes.
        """
        overlap = find_boundaries_overlap(
            self.masks,
            mode="thick",
            connectivity=self.connectivity,
            flag_none=True,
            method=method,
        )

        argmax, valmax = find_masked_extrema(
            self.data,
            list(overlap.values()),
            fill_val=np.nan,
            fill_arg=None,
        )

        # unpack output
        out_arg: dict[tuple[int, int], _ExtremaArg] = {}
        out_max = np.full((self.nfeature,) * 2, dtype=float, fill_value=fill_value)
        if method == "approx":
            for (i, j), arg, val in zip(
                cast("Iterable[tuple[int, int]]", overlap), argmax, valmax
            ):
                out_max[i, j] = out_max[j, i] = val
                out_arg[i, j] = arg

        elif method == "exact":
            # attach keys to argmax, valmax
            overlap_keys = list(overlap)
            argmax_dict = dict(zip(overlap_keys, argmax))
            valmax_dict = dict(zip(overlap_keys, valmax))

            # loop over unique keys
            for i, j in {
                (i, j) for i, j, _ in cast("Iterable[tuple[int, int, int]]", overlap)
            }:
                vals = [valmax_dict[i, j, index] for index in range(2)]
                # take min value of maxes
                if np.all(np.isnan(vals)):
                    out_arg[i, j] = None
                else:
                    idx_min = np.nanargmin(vals)
                    out_arg[i, j] = argmax_dict[i, j, idx_min]  # type: ignore[index]
                    out_max[i, j] = out_max[j, i] = valmax_dict[i, j, idx_min]  # type: ignore[index]
        return out_arg, out_max

    @property
    def w_min(self) -> MyNDArray:
        """Minimum value of `w` (max `lnPi`) in each phase/region."""
        return -self._data_max[1][:, None]

    @property
    def w_argmin(self) -> list[_ExtremaArg]:
        """Locations of the minimum of `w` in each phase/region"""
        return self._data_max[0]

    @property
    def w_tran(self) -> MyNDArray:
        """
        Minimum value of `w` (max `lnPi`) in the boundary between phases.

        `w_tran[i, j]` is the transition energy between phases `i` and `j`.
        """
        return np.nan_to_num(-self._boundary_max()[1], nan=np.inf)

    @property
    def w_argtran(self) -> dict[tuple[int, int], _ExtremaArg]:
        """Location of `w_tran`"""
        return self._boundary_max()[0]

    @cached.prop
    def delta_w(self) -> MyNDArray:
        """Transition energy ``delta_w[i, j] = w_tran[i, j] - w_min[i]``."""
        return self.w_tran - self.w_min

    def merge_regions(
        self,
        nfeature_max: int | None = None,
        efac: float = 1.0,
        force: bool = True,
        convention: MaskConvention = "image",
        warn: bool = True,
    ) -> tuple[Sequence[MyNDArray], MyNDArray, MyNDArray]:
        """
        Merge labels where free energy energy barrier < efac.

        Interface to :func:`merge_regions`

        Parameters
        ----------
        nfeature_max : int
            maximum number of features/phases to allow
        efac : float, default=0.5
            Energy difference to merge on. When ``w_trans[i, j] - w_min[i] < efac``, phases
            ``i`` and ``j`` will be merged together.
        force : bool, default=True
            if True, then keep going until nfeature <= nfeature_max
            even if min_val > efac.
        {mask_convention}
        warn : bool, default=True
            if True, give warning messages

        Returns
        -------
        masks : list of ndarray of bool
            output masks using `convention`
        w_trans : array
            transition energy for new masks
        w_min : array
            free energy minima of new masks
        """

        return merge_regions(
            w_tran=self.w_tran,
            w_min=self.w_min,
            masks=self.masks,
            nfeature_max=nfeature_max,
            efac=efac,
            force=force,
            convention=convention,
            warn=warn,
        )


def _get_w_data(index: pd.MultiIndex, w: wFreeEnergy) -> dict[str, pd.Series[Any]]:
    w_min = pd.Series(w.w_min[:, 0], index=index, name="w_min")
    w_argmin = pd.Series(w.w_argmin, index=w_min.index, name="w_argmin")

    w_tran = (
        pd.DataFrame(  # type: ignore[call-overload]  # noqa: PD013
            w.w_tran,
            index=index,
            columns=index.get_level_values("phase").rename("phase_nebr"),
        )
        .stack()
        .rename("w_tran")  # pyright: ignore[reportArgumentType]
    )

    # get argtrans values for each index
    index_map = {idx: i for i, idx in enumerate(index.get_level_values("phase"))}
    v = w.w_argtran

    argtran = []
    for idxs in zip(
        *[w_tran.index.get_level_values(_) for _ in ["phase", "phase_nebr"]]
    ):
        i, j = (index_map[_] for _ in idxs)

        if (i, j) in v:
            val = v[i, j]
        elif (j, i) in v:
            val = v[j, i]
        else:
            val = None
        argtran.append(val)

    w_argtran = pd.Series(argtran, index=w_tran.index, name="w_argtran")

    return {
        "w_min": w_min,
        "w_tran": w_tran,
        "w_argmin": w_argmin,
        "w_argtran": w_argtran,
    }  # [index_map, w.w_argtran]}


class wFreeEnergyCollection:  # noqa: N801
    r"""
    Calculate the transition free energies for a :class:`lnpy.lnpiseries.lnPiCollection`.

    :math:`w(N) = \beta f(N) = - \ln \Pi(N)`

    Parameters
    ----------
    parent : lnPiCollection

    Notes
    -----
    An instance of :class:`wFreeEnergyCollection` is normally created from the accessor :meth:`lnpy.lnpiseries.lnPiCollection.wfe`
    """

    def __init__(self, parent: lnPiCollection) -> None:
        self._parent = parent
        self._use_joblib = getattr(self._parent, "_use_joblib", False)

        self._cache: dict[str, Any] = {}

    def _get_items_ws(self) -> tuple[list[pd.Index[Any]], list[wFreeEnergy]]:
        indexes = []
        ws = []
        for _, phases in self._parent.groupby_allbut("phase"):
            indexes.append(phases.index)
            masks = [x.mask for x in phases.to_numpy()]

            ws.append(
                wFreeEnergy(data=phases.iloc[0].data, masks=masks, convention=False)
            )
        return indexes, ws

    @cached.prop
    def _data(self) -> dict[str, pd.Series[Any]]:
        indexes, ws = self._get_items_ws()
        seq = get_tqdm(zip(indexes, ws), total=len(ws), desc="wFreeEnergyCollection")
        out = parallel_map_func_starargs(
            _get_w_data, items=seq, use_joblib=self._use_joblib, total=len(ws)
        )

        return {key: pd.concat([x[key] for x in out]) for key in out[0]}

    @property
    def w_min(self) -> pd.Series[Any]:
        """Minimum energy (maximum `lnPi`) for a given region/phase"""
        return self._data["w_min"]

    @property
    def w_tran(self) -> pd.Series[Any]:
        """Minimum energy (maximum `lnPi`) at boundary between phases"""
        return self._data["w_tran"]

    @property
    def w_argmin(self) -> pd.Series[Any]:
        """Location of :attr:`w_min`"""
        return self._data["w_argmin"]

    @property
    def w_argtran(self) -> pd.Series[Any]:
        """Location of :attr:`w_tran`"""
        return self._data["w_argtran"]

    @property
    def dw(self) -> pd.Series[Any]:
        """Series representation of `dw = w_tran - w_min`"""
        return (self.w_tran - self.w_min).rename("delta_w")

    @property
    def dwx(self) -> xr.DataArray:
        """:mod:`xarray` representation of :attr:`dw`"""
        return self.dw.to_xarray()

    @docfiller.decorate
    def get_dwx(
        self, idx: int, idx_nebr: int | list[int] | None = None
    ) -> xr.DataArray:
        """
        Helper function to get the change in energy from
        phase idx to idx_nebr.

        Parameters
        ----------
        {energy_idx}
        {energy_idx_nebr}

        Returns
        -------
        dw : DataArray
            Transition energy from `idx` to `idx_nebr`
            - if only phase idx exists, dw = np.inf
            - if idx does not exists, dw = 0.0 (no barrier between idx and anything else)
            - else min of transition for idx to idx_nebr
        """

        delta_w = self.dwx

        # reindex so that has idx in phase
        reindex = delta_w.indexes["phase"].union(pd.Index([idx], name="phase"))
        delta_w = delta_w.reindex(phase=reindex, phase_nebr=reindex)

        # much simpler
        if idx_nebr is None:
            delta_w = delta_w.sel(phase=idx)
        else:
            if not isinstance(idx_nebr, list):
                idx_nebr = [idx_nebr]
            if idx not in idx_nebr:
                idx_nebr.append(idx)
            nebrs = delta_w.indexes["phase_nebr"].intersection(idx_nebr)
            delta_w = delta_w.sel(phase=idx, phase_nebr=nebrs)

        return delta_w.min("phase_nebr").fillna(0.0)

    def get_dw(
        self, idx: int, idx_nebr: int | list[int] | None = None
    ) -> pd.Series[Any]:
        """Series version of :meth:`get_dwx`"""
        return self.get_dwx(idx, idx_nebr).to_series()


# @lnPiCollection.decorate_accessor("wfe_phases")
class wFreeEnergyPhases(wFreeEnergyCollection):  # noqa: N801
    """
    Stripped down version of :class:`wFreeEnergyCollection` for single phase grouping.

    This should be used for a collection of lnPi that is at a single state point, with multiple phases.

    Parameters
    ----------
    parent : lnPiCollection

    Notes
    -----
    This is accessed through :attr:`lnpy.lnpiseries.lnPiCollection.wfe_phases`

    """

    @property
    @cached.meth
    def dwx(self) -> xr.DataArray:
        index = list(self._parent.index.get_level_values("phase"))
        masks = [x.mask for x in self._parent]
        w = wFreeEnergy(data=self._parent.iloc[0].data, masks=masks, convention=False)

        dw = w.w_tran - w.w_min
        dims = ["phase", "phase_nebr"]
        coords = dict(zip(dims, [index] * 2))
        return xr.DataArray(dw, dims=dims, coords=coords)

    @property
    @cached.meth
    def dw(self) -> pd.Series[Any]:
        """Series representation of delta_w"""
        return self.dwx.to_series()

    def get_dw(  # type: ignore[override]
        self, idx: int, idx_nebr: int | Iterable[int] | None = None
    ) -> float | MyNDArray:
        dw = self.dwx
        index = dw.indexes["phase"]

        if idx not in index:
            return 0.0

        if idx_nebr is None:
            nebrs = index.drop(idx)
        else:
            if isinstance(idx_nebr, int):
                idx_nebr = [idx_nebr]
            nebrs = [x for x in idx_nebr if x in index]

        if len(nebrs) == 0:
            return np.inf
        return dw.sel(phase=idx, phase_nebr=nebrs).min("phase_nebr").to_numpy()


# @lnPiCollection.decorate_accessor("wfe")
# def wfe_accessor(parent: lnPiCollection) -> wFreeEnergyCollection:
#     """Accessor to :class:`~lnpy.lnpienergy.wFreeEnergyCollection` from `self.wfe`."""
#     return wFreeEnergyCollection(parent)


# @lnPiCollection.decorate_accessor("wfe_phases")
# def wfe_phases_accessor(parent: lnPiCollection) -> wFreeEnergyPhases:
#     """Accessor to :class:`~lnpy.lnpienergy.wFreeEnergyPhases` from `self.wfe_phases`."""
#     return wFreeEnergyPhases(parent)


# create alias accessors
# @lnPiCollection.decorate_accessor("wlnPi")
# def wlnPi_accessor(parent: lnPiCollection) -> wFreeEnergyCollection:
#     """
#     Deprecated accessor to :class:`~lnpy.lnpienergy.wFreeEnergyCollection` from `self.wlnPi`.

#     Alias to `self.wfe`
#     """
#     warn("Using `wlnPi` accessor is deprecated.  Please use `wfe` accessor instead")
#     return parent.wfe


# @lnPiCollection.decorate_accessor("wlnPi_single")
# def wlnPi_single_accessor(parent: lnPiCollection) -> wFreeEnergyPhases:
#     """
#     Deprecated accessor to :class:`~lnpy.lnpienergy.wFreeEnergyPhases` from `self.wlnPi_single`.

#     Alias to `self.wfe_single`
#     """
#     warn("Using `wlnPi_single is deprecated.  Please use `self.wfe_phases` instead")
#     return parent.wfe_phases



import numpy as np
import pandas as pd
import pytest

from lnpy.combine import OverlapError, check_windows_overlap, combine_lnpi


def test_check_windows_overlap() -> None:
    # no overlaps:
    overlap = pd.DataFrame(
        {
            "n": [],
            "window": [],
        }
    )

    with pytest.raises(OverlapError):
        check_windows_overlap(
            overlap,
            windows=range(3),
            state_names="n",
            window_name="window",
            verbose=False,
        )

    overlap = pd.DataFrame({"n": [0, 0, 1, 1, 2, 2], "window": [0, 1, 1, 2, 2, 3]})

    check_windows_overlap(
        overlap,
        windows=range(4),
        state_names="n",
        window_name="window",
        verbose=True,
    )

    # missing node
    with pytest.raises(OverlapError):
        check_windows_overlap(
            overlap,
            windows=range(5),
            state_names="n",
            window_name="window",
            verbose=True,
        )
    # two graphs
    overlap = pd.DataFrame(
        {
            "n": [0, 0, 1, 1],
            "window": [0, 1, 2, 3],
        }
    )

    with pytest.raises(OverlapError):
        check_windows_overlap(
            overlap,
            windows=range(4),
            state_names="n",
            window_name="window",
            verbose=True,
        )


@pytest.mark.parametrize("check_connected", [False, True])
@pytest.mark.parametrize("use_sparse", [False, True])
def test_simple(check_connected: bool, use_sparse: bool) -> None:
    x = np.linspace(0, 10)

    table = pd.DataFrame(
        {
            "x": x,
            "y": np.sin(x),
        }
    )

    mid = len(table) // 2

    # no overlap:
    dfs = [table.iloc[:mid], table.iloc[mid:]]
    with pytest.raises(OverlapError):
        combine_lnpi(
            dfs,
            state_names="x",
            lnpi_name="y",
            use_sparse=use_sparse,
            check_connected=check_connected,
        ).groupby("x", as_index=False).mean()

    # single table
    dfs = [table]
    new = (
        combine_lnpi(
            dfs,
            state_names="x",
            lnpi_name="y",
            use_sparse=use_sparse,
            check_connected=check_connected,
        )
        .groupby("x", as_index=False)
        .mean()
    )

    pd.testing.assert_frame_equal(table, new[["x", "y"]])

    dfs = [table.iloc[: mid + 1], table.iloc[mid:].assign(y=lambda x: x["y"] + 10)]
    new = (
        combine_lnpi(
            dfs,
            state_names="x",
            lnpi_name="y",
            use_sparse=use_sparse,
            check_connected=check_connected,
        )
        .groupby("x", as_index=False)
        .mean()
    )

    np.testing.assert_allclose(table["x"], new["x"])
    np.testing.assert_allclose(table["y"], new["y"])

    # with window name already there
    dfs = [x.assign(window="hello", other=1) for x in dfs]
    new = (
        combine_lnpi(
            dfs,
            state_names="x",
            lnpi_name="y",
            use_sparse=use_sparse,
            check_connected=check_connected,
        )
        .groupby("x", as_index=False)
        .mean()
    )

    np.testing.assert_allclose(table["x"], new["x"])
    np.testing.assert_allclose(table["y"], new["y"])

    # multiple splits:
    n = 5
    split = len(table) // n
    lb = 0

    dfs = []
    for i in range(n + 1):
        ub = lb + split + 1
        dfs.append(table.iloc[lb:ub].assign(y=lambda x: x["y"] + 10 * i))  # noqa: B023

        lb = lb + split
        if lb >= len(table):
            break

    new = (
        combine_lnpi(
            dfs,
            state_names="x",
            lnpi_name="y",
            use_sparse=use_sparse,
            check_connected=check_connected,
        )
        .groupby("x", as_index=False)
        .mean()
    )

    np.testing.assert_allclose(table["x"], new["x"])
    np.testing.assert_allclose(table["y"], new["y"])



"""Version compatibility code should go here."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Hashable, Iterable

    from scipy.optimize import RootResults  # pyright: ignore[reportMissingTypeStubs]

    from ._typing import MyNDArray


def xr_dot(
    *arrays: Any,
    dim: str | Iterable[Hashable] | "ellipsis" | None = None,  # noqa: F821, UP037
    **kwargs: Any,
) -> Any:
    """
    Interface to xarray.dot.

    Remove a deprecation warning for older xarray versions.
    xarray deprecated `dims` keyword.  Use `dim` instead.
    """
    import xarray as xr

    try:
        return xr.dot(*arrays, dim=dim, **kwargs)  # type: ignore[arg-type,unused-ignore]
    except TypeError:
        return xr.dot(*arrays, dims=dim, **kwargs)  # type: ignore[arg-type,unused-ignore]


def rootresults(
    root: float | MyNDArray | None,
    iterations: int,
    function_calls: int,
    flag: int,
    method: str = "User",
) -> RootResults:
    """
    Create :class:`scipy.optimize.RootResults` object.

    There are differences in `RootResults` for different
    versions of scipy, so have this as an interface
    """
    from scipy.optimize import RootResults  # pyright: ignore[reportMissingTypeStubs]

    try:
        return RootResults(
            root=root,
            iterations=iterations,
            function_calls=function_calls,
            flag=flag,
            method=method,  # pyright: ignore[reportCallIssue]
        )
    except TypeError:
        return RootResults(  # pyright: ignore[reportCallIssue]
            root=root,
            iterations=iterations,
            function_calls=function_calls,
            flag=flag,
        )



"""Publicly accessible classes/routines."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import ensembles, lnpienergy, segment  # noqa: TCH004
    from .combine import combine_lnpi  # noqa: TCH004
    from .lnpidata import lnPiMasked  # noqa: TCH004
    from .lnpiseries import lnPiCollection  # noqa: TCH004
    from .options import OPTIONS, set_options  # noqa: TCH004
    from .segment import PhaseCreator  # noqa: TCH004
else:
    import lazy_loader as lazy

    __getattr__, __dir__, _ = lazy.attach(
        __name__,
        submodules=["ensembles", "lnpienergy", "segment"],
        submod_attrs={
            "lnpidata": ["lnPiMasked"],
            "lnpiseries": ["lnPiCollection"],
            "options": ["OPTIONS", "set_options"],
            "segment": ["PhaseCreator"],
        },
    )


# updated versioning scheme
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

try:
    __version__ = _version("tmmc-lnpy")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "999"

__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "OPTIONS",
    "PhaseCreator",
    "__author__",
    "__email__",
    "__version__",
    "combine_lnpi",
    "ensembles",
    "lnPiCollection",
    "lnPiMasked",
    "lnpienergy",
    "segment",
    "set_options",
]



"""Common docstrings."""

from module_utilities.docfiller import DocFiller

_docstrings = """
Parameters
----------
copy : bool, optional
    If True, copy the data.  If False, attempt to use view.
copy_kws : mapping, optional
    extra arguments to copy
lnz : float or sequence of float
    Value(s) of ``lnz`` (log of activity)
state_kws : Mapping, optional
    'State' variables.  Common key/value pairs used by other methods:

    * beta (float) : inverse temperature
    * volume (float) : volume
extra_kws : Mapping, Optional
    Extra parameters.  These are passed along to new objects without modification.
    Common key/value pairs:

* PE (array-like) : value of total potential energy.
    Used in calculation of canonical and grand canonical average potential energy.
fill_value : scalar, default=nan
    Value to fill for masked elements.
data : array-like
    Value of lnPi.  Number of dimensions should be same as ``len(lnz)``.
    For single component, ``data[k]`` is the value of lnPi for `k` particles.
mask_masked | mask : None or ndarray of bool
    Mask using "masked" convention. Where `mask` is `True`, values are excluded.
mask_image | mask : None or ndarray of bool
    Mask using "image" convention. Where `mask` is `True`, values are included.
masks_masked | masks : sequence of None or ndarray of bool
    Masks using "masked" convention.  Where `mask[i]` is `True`, values are excluded
    for sample `i`.
masks_image | masks : sequence of None or ndarray of bool
    Masks using "image" convention.  Where `masks[i]` is `True`, values are included
    for sample `i`.
masks_general | masks: sequence of None or ndarray of bool
    Masking arrays.
mask_general | masks: None or ndarray of bool
    Masking array.
mask_convention | convention : string or bool
    Convention for mask.  Allowable values are:

    * 'image' or True : `True` values included, `False` values excluded.
        This is the normal convention in :mod:`scipy.ndimage`.
    * 'masked' or False:  `False` values are included, `True` values are excluded.
        This is the convention in :mod:`numpy.ma`
labels : np.ndarray of int
    Where `labels == k`, indicates regions where data corresponds to the kth label.
base : object
    lnPi data object.
peak_kws : mapping, optional
    Optional parameters to :func:`~lnpy.segment.peak_local_max_adaptive`
watershed_kws : mapping, optional
    Optional parameters to :func:`~skimage.segmentation.watershed`
features : sequence of int
    If specified, extract only those locations where ``labels == feature``
    for all values ``feature in features``.  That is, select a subset of
    unique label values.
include_boundary : bool
    if True, include boundary regions in output mask
check_features : bool
    if True, then make sure each feature is in labels
labels : ndarray of int
    Each unique value `i` in `labels` indicates a mask.
    That is ``labels == i`` is a mask for feature `i`.
ffill : bool, default=True
    Do forward filling
bfill : bool, default=False
    Do back filling
fill_limit | limit : int, default None
    The maximum number of consecutive NaN values to forward fill. In
    other words, if there is a gap with more than this number of
    consecutive NaNs, it will only be partially filled. Must be greater
    than 0 or None for no limit.
find_boundary_connectivity  | connectivity: int, optional
    Defaults to ``masks[0].ndim``
energy_idx | idx : int
    phase index to consider transitions from
energy_idx_nebr | idx_nebr : int or list, optional
    if supplied, consider transition from idx to idx_nebr or minimum of all element in idx_nebr.
    Default behavior is to return minimum transition from idx to all other neighboring regions
"""


_accessor_docs = {
    "xge": "Accessor to :class:`~lnpy.ensembles.xGrandCanonical`.",
    "xce": "Accessor to :class:`~lnpy.ensembles.xCanonical`.",
}


docfiller = DocFiller.from_docstring(_docstrings, combine_keys="parameters").update(
    accessor=_accessor_docs
)



"""
Examples (:mod:`~lnpy.examples`)
================================
"""

from __future__ import annotations

try:
    import importlib_resources as resources
except ImportError:
    from importlib import resources  # type: ignore[no-redef]

import json
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, TypedDict

import numpy as np
import xarray as xr

# from .lnpiseries import lnPiCollection
from .segment import BuildPhasesBase, PhaseCreator
from .utils import dataset_to_lnpimasked

if TYPE_CHECKING:
    from typing import Any, Iterator, Literal, Sequence

    from ._typing import MyNDArray
    from .lnpidata import lnPiMasked

    _ExampleNames = Literal["lj_sub", "lj_sup", "ljmix_sup", "hsmix", "watermof"]


def json_to_dict(basename: str) -> dict[str, Any]:
    """
    Load a json file into a dict.

    All files names are relative to 'lnpy/data/'

    Parameters
    ----------
    basename : string
        Name of file to load in 'lnpy/data' directory.

    Returns
    -------
    output : dict

    """

    if basename.endswith(".gz"):
        import gzip

        fopen = gzip.open
    else:
        fopen = open  # type: ignore[assignment]

    with fopen(resources.files("lnpy.data").joinpath(basename), "r") as f:  # pyright: ignore[reportCallIssue,reportArgumentType]
        return json.load(f)  # type: ignore[no-any-return]


class ExampleDict(TypedDict):
    """Example dict"""

    lnPi_data: MyNDArray
    lnPi_mask: MyNDArray
    state_kws: dict[str, Any]
    extra_kws: dict[str, Any]
    lnz: MyNDArray


def load_example_dict(name: _ExampleNames) -> ExampleDict:
    """
    Load a dictionary of data

    Parameters
    ----------
    name : {'lj_sub', 'lj_sup', 'ljmix_sup', 'hsmix', 'watermof}
    """

    ref = load_example_lnpimasked(name)

    return ExampleDict(
        lnPi_data=ref.data,
        lnPi_mask=ref.mask,
        state_kws=ref.state_kws,
        extra_kws=ref.extra_kws,
        lnz=ref.lnz,
    )


def load_example_lnpimasked(name: _ExampleNames) -> lnPiMasked:
    """
    Load an example file

    Parameters
    ----------
    name : {'lj_sub', 'lj_sup', 'ljmix_sup', 'hsmix', 'watermof}
    """
    extensions = {
        "lj_sub": "json",
        "lj_sup": "json",
        "ljmix_sup": "json.gz",
        "hsmix": "json.gz",
        "watermof": "json",
    }

    if name not in extensions:
        msg = f"{name=} not in {extensions=}"
        raise ValueError(msg)

    basename = f"{name}_example.{extensions[name]}"

    d = json_to_dict(basename)
    ds = xr.Dataset.from_dict(d)

    return dataset_to_lnpimasked(ds)


@dataclass
class Example:
    """Dataclass to hold example data."""

    #: Reference state :class:`~lnpy.lnpidata.lnPiMasked`.
    ref: lnPiMasked
    #: :class:`~lnpy.segment.PhaseCreator` instance
    phase_creator: PhaseCreator
    #: Callable to build phases.
    build_phases: BuildPhasesBase | None

    def to_dict(self) -> dict[str, Any]:
        """Transform class to dictionary."""
        return asdict(self)

    def unpack(self, keys: list[str] | None = None) -> Iterator[Any]:
        """Unpack keys."""
        if keys is None:
            keys = ["ref", "phase_creator", "build_phases"]
        return (getattr(self, k) for k in keys)


def lj_sup_example() -> Example:
    """Create an :class:`Example` instance for a Lennard-Jones fluid (subcritical)"""
    ref = load_example_lnpimasked("lj_sup")

    phase_creator = PhaseCreator(
        nmax=1,
        nmax_peak=1,
        ref=ref,
        merge_kws={"efac": 0.8},
        tag_phases=None,
    )

    build_phases = phase_creator.build_phases_mu([None])

    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def tag_phases_single_comp_simple(x: Sequence[lnPiMasked]) -> MyNDArray:
    if len(x) > 2:
        msg = "bad tag function"
        raise ValueError(msg)
    argmax0 = np.array([xx.local_argmax()[0] for xx in x])
    return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)


def lj_sub_example() -> Example:
    """Create an :class:`Example` instance for a Lennard-Jones fluid (subcritical)"""
    ref = load_example_lnpimasked("lj_sub")

    phase_creator = PhaseCreator(
        nmax=2,
        nmax_peak=4,
        ref=ref,
        merge_kws={"efac": 0.8},
        tag_phases=tag_phases_single_comp_simple,
    )

    build_phases = phase_creator.build_phases_mu([None])

    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def ljmix_sup_example() -> Example:
    """Create an :class:`Example` instance for a Lennard-Jones mixture (supercritical)."""
    ref = load_example_lnpimasked("ljmix_sup")

    phase_creator = PhaseCreator(
        nmax=1,
        ref=ref,
    )

    build_phases = phase_creator.build_phases_mu([None])
    return Example(ref=ref, phase_creator=phase_creator, build_phases=build_phases)


def hsmix_example() -> Example:
    """Create an :class:`Example` instance for a hard-sphere mixture."""

    def tag_phases(x: Sequence[lnPiMasked]) -> MyNDArray:
        if len(x) > 2:
            msg = "bad tag function"
            raise ValueError(msg)
        argmax0 = np.array([xx.local_argmax()[0] for xx in x])
        return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)

    ref = load_example_lnpimasked("hsmix")

    phase_creator = PhaseCreator(
        nmax=2, nmax_peak=4, ref=ref, tag_phases=tag_phases, merge_kws={"efac": 0.8}
    )

    return Example(ref=ref, phase_creator=phase_creator, build_phases=None)


# def watermof_example():
#     def tag_phases(x):
#         if len(x) > 2:
#             raise ValueError("bad tag function")
#         argmax0 = np.array([xx.local_argmax()[0] for xx in x])
#         return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)

#     ref = load_example_lnpimasked('watermof')

#     phase_creator = PhaseCreator(
#         nmax=2,
#         nmax_peak=10,
#         merge_kws={'efac': 0.8}

#     )



"""Config for doctests and test collection"""

import pytest


@pytest.fixture(autouse=True)
def add_standard_imports(doctest_namespace) -> None:  # noqa: ARG001
    import numpy as np
    # import pandas as pd
    # import xarray as xr

    # doctest_namespace["np"] = np
    # doctest_namespace["pd"] = pd
    # doctest_namespace["xr"] = xr

    np.set_printoptions(precision=4)


def pytest_ignore_collect(collection_path, path, config) -> None:  # noqa: ARG001
    import sys

    if sys.version_info < (3, 9):
        return "tests/test_" not in str(collection_path)

    return False



"""
Options (:mod:`~lnpy.options`)
==============================
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypedDict, cast

if TYPE_CHECKING:
    from types import TracebackType

ValidatorFunc = Callable[[Any], bool]


class Options(TypedDict, total=False):
    """Options."""

    tqdm_use: bool
    tqdm_len_calc: int
    tqdm_len_build: int
    tqdm_leave: bool
    tqdm_bar: str

    joblib_use: bool
    joblib_n_jobs: int
    joblib_backend: str | None
    joblib_kws: dict[str, Any]
    joblib_len_calc: int
    joblib_len_build: int


class OptionsReq(TypedDict, total=True):
    """Options with required parameters."""

    tqdm_use: bool
    tqdm_len_calc: int
    tqdm_len_build: int
    tqdm_leave: bool
    tqdm_bar: str

    joblib_use: bool
    joblib_n_jobs: int
    joblib_backend: str | None
    joblib_kws: dict[str, Any]
    joblib_len_calc: int
    joblib_len_build: int


class Validators(TypedDict):
    """Validators."""

    tqdm_use: ValidatorFunc
    tqdm_len_calc: ValidatorFunc
    tqdm_len_build: ValidatorFunc
    tqdm_leave: ValidatorFunc
    tqdm_bar: ValidatorFunc

    joblib_use: ValidatorFunc
    joblib_n_jobs: ValidatorFunc
    joblib_backend: ValidatorFunc
    joblib_kws: ValidatorFunc
    joblib_len_calc: ValidatorFunc
    joblib_len_build: ValidatorFunc


OPTIONS: OptionsReq = {
    "tqdm_use": True,
    "tqdm_len_calc": 100,
    "tqdm_len_build": 500,
    "tqdm_leave": False,
    "tqdm_bar": "default",
    "joblib_use": True,
    "joblib_n_jobs": -1,
    "joblib_backend": None,
    "joblib_kws": {},
    "joblib_len_calc": 200,
    "joblib_len_build": 500,
}


def _isbool(x: Any) -> bool:
    return isinstance(x, bool)


def _isint(x: Any) -> bool:
    return isinstance(x, int)


def _isstr(x: Any) -> bool:
    return isinstance(x, str)


def _isdict(x: Any) -> bool:
    return isinstance(x, dict)


def _isstr_or_none(x: Any) -> bool:
    return x is None or _isstr(x)


_VALIDATORS: Validators = {
    "tqdm_use": _isbool,
    "tqdm_len_calc": _isint,
    "tqdm_len_build": _isint,
    "tqdm_leave": _isbool,
    "tqdm_bar": lambda x: x in {"default", "text", "notebook"},
    "joblib_use": _isbool,
    "joblib_n_jobs": _isint,
    "joblib_backend": _isstr_or_none,
    "joblib_kws": _isdict,
    "joblib_len_calc": _isint,
    "joblib_len_build": _isint,
}

_SETTERS: dict[str, Any] = {}


def _apply_update(options_dict: Options) -> None:
    for k, v in options_dict.items():
        if k in _SETTERS:
            _SETTERS[k](v)
    OPTIONS.update(options_dict)


class set_options:  # noqa: N801
    """
    Set options for xarray in a controlled context.
    Currently supported options:

    * `tqdm_use` : if `True`, use progress bar where appropriate
    * `tqdm_len_calc` : min length for using bar in calculations of properties
    * `tqdm_len_build` : min length for building Collection objects
    * `tqdm_leave` : if True, leave bar after execution.  Default=False
    * `joblib_use` : if `True`, use joblib
    * `joblib_n_jobs` : number of processors to use, default=all processors
    * `joblib_backend` : backend to use.  Default='multiprocessing'.
    * `joblib_kws` : dictionary of arguments to joblib.Parallel.
    * `joblib_len_calc` : min length to use parallel in calculations
    * `joblib_len_build` : min length to use parallel in build

    Examples
    --------
    You can use ``set_options`` either as a context manager:

    >>> import lnpy
    >>> with lnpy.set_options(use_tqdm=True, tqdm_min_len_calc=50):  # doctest: +SKIP
    ...     c.xge.betaOmega()


    Or to set global options:

    >>> _ = lnpy.set_options(tqdm_len_calc=50)
    """

    def __init__(self, **kwargs: Any) -> None:
        self.old: Options = {}
        for k, v in cast(Options, kwargs).items():
            if k not in OPTIONS:
                msg = f"argument name {k!r} is not in the set of valid options {set(OPTIONS)!r}"
                raise ValueError(msg)
            if k in _VALIDATORS and not _VALIDATORS[k](v):  # type: ignore[literal-required]
                msg = f"option {k!r} given an invalid value: {v!r}"
                raise ValueError(msg)
            self.old[k] = OPTIONS[k]  # type: ignore[literal-required]
        _apply_update(cast(Options, kwargs))

    def __enter__(self) -> None:
        return

    def __exit__(
        self,
        type: type[BaseException] | None,  # noqa: A002
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        _apply_update(self.old)



"""
Typing definitions for :mod:`lnpy`
==================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence, TypeVar, Union

import xarray as xr
from numpy.typing import ArrayLike, NDArray

from ._typing_compat import Concatenate, ParamSpec, TypeAlias

if TYPE_CHECKING:
    import numpy as np

    # Note: use fully qualified names
    import lnpy.lnpidata
    import lnpy.lnpiseries

    from .ensembles import xCanonical, xGrandCanonical


__all__ = [
    "C_Ensemble",
    "F",
    "FuncType",
    "IndexingInt",
    "MaskConvention",
    "MyNDArray",
    "P",
    "PeakError",
    "PeakStyle",
    "PhasesFactorySignature",
    "R",
    "T",
    "T_Ensemble",
    "TagPhasesSignature",
    "xArrayLike",
]


T_Ensemble = TypeVar("T_Ensemble", "xGrandCanonical", "xCanonical")
"""TypeVar for Ensemble."""

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")
C_Ensemble: TypeAlias = Callable[Concatenate[T_Ensemble, P], R]

FuncType = Callable[..., Any]

F = TypeVar("F", bound=FuncType)


MyNDArray: TypeAlias = NDArray[Any]
"""Alias for simple :class:`numpy.typing.NDArray`"""


IndexingInt: TypeAlias = Union[
    int,
    "np.int_",
    "np.integer[Any]",
    "np.unsignedinteger[Any]",
    "np.signedinteger[Any]",
    "np.int8",
]


xArrayLike: TypeAlias = Union[ArrayLike, xr.DataArray]  # noqa: N816

IndexIterScalar: TypeAlias = Union[str, bytes, bool, int, float]
Scalar: TypeAlias = IndexIterScalar

# Segmentation stuff
PeakStyle = Literal["indices", "mask", "marker"]
PeakError = Literal["ignore", "raise", "warn"]


TagPhasesSignature = Callable[
    [Sequence["lnpy.lnpidata.lnPiMasked"]], Union[Sequence[int], MyNDArray]
]
"""Signature for tag_phases function."""

PhasesFactorySignature = Callable[..., "lnpy.lnpiseries.lnPiCollection"]
"""Signature for phases_factory function."""


MaskConvention = Literal["image", "masked", True, False]
"""Convention for boolean masks."""



"""
Thermodynamic stability (:mod:`~lnpy.stability`)
================================================

Calculation of spinodal and binodal
"""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, overload

import numpy as np
from module_utilities import cached

from ._compat import rootresults
from .utils import RootResultDict, array_to_scalar, rootresults_to_rootresultdict

if TYPE_CHECKING:
    from typing import Any, Iterable, Literal, Mapping

    from scipy.optimize import RootResults  # pyright: ignore[reportMissingTypeStubs]

    from ._typing import MyNDArray
    from ._typing_compat import Self
    from .lnpidata import lnPiMasked
    from .lnpiseries import lnPiCollection
    from .segment import BuildPhasesBase


class RootResultTotal(RootResultDict, total=False):
    """Modified root result"""

    left: lnPiCollection
    right: lnPiCollection
    left_done: bool
    right_done: bool
    info: str
    bracket_iteration: int
    from_solve: bool


def _rootresults_to_rootresulttotal(
    r: RootResults,
    *,
    left: lnPiCollection | None = None,
    right: lnPiCollection | None = None,
    left_done: bool | None = None,
    right_done: bool | None = None,
    info: str | None = None,
    bracket_iteration: int | None = None,
    from_solve: bool | None = None,
    residual: float | None = None,
) -> RootResultTotal:
    output = RootResultTotal(**rootresults_to_rootresultdict(r, residual=residual))  # type: ignore[typeddict-item]

    if left is not None:
        output["left"] = left
    if right is not None:
        output["right"] = right
    if left_done is not None:
        output["left_done"] = left_done
    if right_done is not None:
        output["right_done"] = right_done
    if info is not None:
        output["info"] = info
    if bracket_iteration is not None:
        output["bracket_iteration"] = bracket_iteration
    if from_solve is not None:
        output["from_solve"] = from_solve

    return output


# ###############################################################################
# Spinodal routines
def _initial_bracket_spinodal_right(
    collection: lnPiCollection,
    build_phases: BuildPhasesBase,
    idx: int,
    idx_nebr: int | None = None,
    efac: float = 1.0,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    # vmax: float = 1e5,
    ntry: int = 20,
    step: int = +1,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
) -> tuple[lnPiCollection, lnPiCollection]:
    """
    Find initial bracketing lnPiCollection for spinodal.

    Parameters
    ----------
    ref : lnPiMasked, optional
    build_phases : callable
        scalar function to build phases
    collection : lnPiCollection
        initial estimates to work from
    idx, idx_nebr : int
        id's of from/to phases.
    lnz_in : list
        list with value of static chem pot, and None for variable. e.g.,
        lnz_in=[None,0.0] implies lnz[0] is variable, and lnz[1]=0.0
    efac : float, default=1.0
        cutoff value for spinodal
    dlnz : float, default=0.5
        factor to kick back if collection doesn't already have left and right bounds
    vmax : float default=1e20
        value indicating no transition, but phase `idx` present.
    ntry : int, default=20
        number of times to try kicking forward/backwards to find bracket
    step : int, default=+1
        if +1, step forward
        if -1, step backward
    build_phases : callable, optional.
        function that returns a Phases object.
    build_kws : dict
        extra arguments to build_phases

    Returns
    -------
    left,right: lnPiCollection
        left and right bracketing :class:`~lnpy.lnpiseries.lnPiCollection` objects

    """

    if build_kws is None:
        build_kws = {}
    if efac <= 0:
        msg = "efac must be positive"
        raise ValueError(msg)

    # Series representation of dw
    s = collection.wfe.get_dw(idx, idx_nebr)
    if step < 0:
        s = s.iloc[-1::-1]

    # see if contain "left" bounds
    left = None
    s_idx = s[s > 0.0]
    if len(s_idx) == 0:
        msg = f"no phase {idx}"
        raise ValueError(msg)

    ss = s_idx[s_idx > efac]
    if len(ss) > 0:
        # get last one
        left = collection.mloc[ss.index[[-1]]]
    else:
        new_lnz = collection.mloc[s_idx.index[[0]]]._get_lnz(build_phases.index)
        for _i in range(ntry):
            new_lnz -= step * dlnz
            t = build_phases(new_lnz, ref=ref, **build_kws)
            if (
                idx in t._get_level("phase")
                and t.wfe_phases.get_dw(idx, idx_nebr) > efac
            ):
                left = t
                break

    if left is None:
        msg = "could not find left"
        raise RuntimeError(msg)

    # right
    right = None
    ss = s[s < efac]
    if len(ss) > 0:
        right = collection.mloc[ss.index[[0]]]
    else:
        new_lnz = collection.mloc[s.index[[-1]]]._get_lnz(build_phases.index)
        dlnz_ = dlnz
        for _i in range(ntry):
            new_lnz += step * dlnz_
            t = build_phases(new_lnz, ref=ref, **build_kws)
            if (
                idx not in t._get_level("phase")
                or t.wfe_phases.get_dw(idx, idx_nebr) < efac
            ):
                right = t
                break
            dlnz_ *= dfac

    if right is None:
        msg = "could not find right"
        raise RuntimeError(msg)
    return left, right


def _refine_bracket_spinodal_right(
    left: lnPiCollection,
    right: lnPiCollection,
    build_phases: BuildPhasesBase,
    idx: int,
    idx_nebr: int | None = None,
    efac: float = 1.0,
    nmax: int = 30,
    vmax: float = 1e5,
    vmin: float = 0.0,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
    close_kws: Mapping[str, Any] | None = None,
) -> tuple[lnPiCollection | None, lnPiCollection | None, RootResultTotal]:
    """
    Find refined bracket.

    Finds where ``efac<wfe<vmax`` and
    ``vmin<wfe<efac``

    Parameters
    ----------
    left, right : lnPiCollection
        left and right initial brackets
    idx, idx_nebr : int
        from/to phase id's
    efac : float, default=1.0
        cutoff value for spinodal
    nmax : int, default=30
        max number of iterations
    vmin, vmax : float
        See above
    build_phases : callable
    build_kws : dict
    close_kwargs : dict
        arguments to :func:`numpy.allclose`

    Returns
    -------
    left,right : lnPiCollection
        left and right phases bracketing spinodal

    r : :class:`scipy.optimize.zeros.RootResults` object
    """
    left_done = False
    right_done = False
    if build_kws is None:
        build_kws = {}
    if close_kws is None:
        close_kws = {}

    for i in range(nmax):
        # if idx in left.index and idx_nebr in left.index:
        # dw = _get_dw(left, idx, idx_nebr)
        dw = left.wfe_phases.get_dw(idx, idx_nebr)
        if dw < vmax and dw > efac:
            left_done = True

        # dw = _get_dw(right, idx, idx_nebr)
        dw = right.wfe_phases.get_dw(idx, idx_nebr)
        if dw > vmin and dw < efac:
            right_done = True

        #########
        # checks
        if left_done and right_done:
            # find bracket
            r = rootresults(root=None, iterations=i, function_calls=i, flag=1)
            return left, right, _rootresults_to_rootresulttotal(r)

        ########
        # converged?
        if np.allclose(left._get_lnz(), right._get_lnz(), **close_kws):
            # we've reached a breaking point
            if left_done:
                # can't find a lower bound to efac, just return where we're at
                r = rootresults(
                    root=left._get_lnz(), iterations=i + 1, function_calls=i, flag=0
                )

                r = _rootresults_to_rootresulttotal(
                    r,
                    left=left,
                    right=right,
                    left_done=left_done,
                    right_done=right_done,
                    info="all close and left_done",
                )
                return left, right, r

            # all close, and no good on either end -> no spinodal
            r = rootresults(root=None, iterations=i + 1, function_calls=i, flag=1)

            r = _rootresults_to_rootresulttotal(
                r,
                left=left,
                right=right,
                left_done=left_done,
                right_done=right_done,
                info="all clase and not left_done",
            )

            return None, None, r

        # mid point phases
        lnz_mid = 0.5 * (
            left._get_lnz(build_phases.index) + right._get_lnz(build_phases.index)
        )

        mid = build_phases(lnz_mid, ref=ref, **build_kws)
        dw = mid.wfe_phases.get_dw(idx, idx_nebr)

        if (idx in mid._get_level("phase")) and (
            mid.wfe_phases.get_dw(idx, idx_nebr) >= efac
        ):
            left = mid
        else:
            right = mid

    msg = f"""
    did not finish
    ntry      : {i}
    idx       : {idx}
    idx_nebr  : {idx_nebr}
    left lnz   : {left._get_lnz()}
    right lnz  : {right._get_lnz()}
    doneleft  : {left_done}
    doneright : {right_done}
    """
    raise RuntimeError(msg)


def _get_step(collection: lnPiCollection, idx: int, idx_nebr: int | None) -> int:
    """
    Find step value on

    if DeltabetaE[-1,ID] - DeltabetaE[0,ID]<0 -> step=+1 (go right)

    else step = -1
    """
    delta = collection.zloc[[-1]].wfe_phases.get_dw(idx, idx_nebr) - collection.zloc[
        [0]
    ].wfe_phases.get_dw(idx, idx_nebr)
    if delta == 0:
        msg = "could not determine step, delta==0"
        raise ValueError(msg)
    if delta < 0.0:
        return +1
    return -1


class _SolveSpinodal:
    collection: lnPiCollection

    def __init__(
        self,
        build_phases: BuildPhasesBase,
        idx: int,
        idx_nebr: int | None = None,
        efac: float = 1.0,
        ref: lnPiMasked | None = None,
        build_kws: Mapping[str, Any] | None = None,
    ) -> None:
        self.build_phases = build_phases
        self.idx = idx
        self.idx_nebr = idx_nebr
        self.efac = efac
        self.ref = ref
        self.build_kws = build_kws or {}

    def objective(self, x: float) -> float:
        self.collection = self.build_phases(x, ref=self.ref, **self.build_kws)
        dw = self.collection.wfe_phases.get_dw(self.idx, self.idx_nebr)

        return array_to_scalar(dw) - self.efac

    def solve(
        self, a: float, b: float, **kws: Any
    ) -> tuple[float, RootResultTotal, lnPiCollection]:
        from scipy.optimize import brentq

        xx, r = brentq(self.objective, a, b, full_output=True, **kws)

        return (
            self.collection._get_lnz(self.build_phases.index),
            _rootresults_to_rootresulttotal(r, residual=self.objective(xx)),
            self.collection,
        )


def get_spinodal(
    collection: lnPiCollection,
    build_phases: BuildPhasesBase,
    idx: int,
    idx_nebr: int | None = None,
    efac: float = 1.0,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    vmin: float = 0.0,
    vmax: float = 1e5,
    ntry: int = 20,
    step: int | None = None,
    nmax: int = 20,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
    close_kws: Mapping[str, Any] | None = None,
    solve_kws: Mapping[str, Any] | None = None,
) -> tuple[lnPiCollection | None, RootResultTotal]:
    """
    Locate spinodal point for a given pair of phases.

    Parameters
    ----------
    ref : lnPiMasked
    collection : lnPiCollection
        initial estimates to work from.  Function assumes collection is in lnz sorted order
    idx, idx_nebr : int
        from/to phase id
    lnz_in : list
        list with value of static chem pot, and None for variable. e.g.,
        lnz_in=[None,0.0] implies lnz[0] is variable, and lnz[1]=0.0
    efac : float, optional
        cutoff value for spinodal
    dlnz : float, optional
        factor to kick back if collection doesn't already have left and right bounds
    vmin : float, optional
        Value denoting ``vmin``, i.e., value of free energy difference phase does not exist.
    vmax : float, optional
        value indicating no transition, but phase `idx` is present.
    ntry : int, default=20
        number of times to try kicking forward/backwards to find bracket
    step : int or None, default=None
        if +1, step forward
        if -1, step backward
        if None, try to determine step
    nmax : int, default=20
        max number of steps to refine bracket
    build_phases : callable, optional
        Function to create Phases.  Default is that from get_default_phasecreator
    build_kws : dict, optional
        extra arguments to ``build_phases``
    close_kws : dict, optional
        arguments to :func:`numpy.allclose`
    solve_kws : dict, optional
        extra arguments to :func:`scipy.optimize.brentq`

    Returns
    -------
    out : :class:`~lnpy.lnpiseries.lnPiCollection` object at spinodal point
    r : object
        Info object

    """
    if len(collection) < 2:
        raise ValueError
    build_kws = build_kws or {}
    close_kws = close_kws or {}
    solve_kws = solve_kws or {}

    step = step or _get_step(collection, idx=idx, idx_nebr=idx_nebr)

    if step not in {-1, +1}:
        msg = f"{step=} must by +/- 1"
        raise ValueError(msg)

    # get initial bracket
    left_initial, right_initial = _initial_bracket_spinodal_right(
        collection,
        idx=idx,
        idx_nebr=idx_nebr,
        efac=efac,
        dlnz=dlnz,
        dfac=dfac,
        # vmax=vmax,
        ntry=ntry,
        step=step,
        ref=ref,
        build_phases=build_phases,
        build_kws=build_kws,
    )

    left, right, rr = _refine_bracket_spinodal_right(
        left_initial,
        right_initial,
        idx=idx,
        idx_nebr=idx_nebr,
        efac=efac,
        nmax=nmax,
        vmin=vmin,
        vmax=vmax,
        ref=ref,
        build_phases=build_phases,
        build_kws=build_kws,
        close_kws=close_kws,
    )

    if left is None and right is None:
        # no spinodal found and left and right are close
        spin = None
        r = rr
    elif rr["converged"]:
        # converged to a solution
        spin = left
        r = rr
        r["bracket_iteration"] = rr["iterations"]
        r["from_solve"] = False
    else:
        # assert left is not None
        # assert right is not None
        if left is None or right is None:
            msg = f"{left=} and {right=} cannot be None"
            raise ValueError(msg)
        # solve
        if step == -1:
            left, right = right, left

        a, b = (x._get_lnz(build_phases.index) for x in [left, right])
        _lnz, r, spin = _SolveSpinodal(
            ref=ref,
            idx=idx,
            idx_nebr=idx_nebr,
            efac=efac,
            build_phases=build_phases,
            build_kws=build_kws,
        ).solve(a=a, b=b, **solve_kws)

        r["bracket_iteration"] = rr["iterations"]
        r["from_solve"] = True

    return spin, r


################################################################################
# Binodal routines
class _SolveBinodal:
    """
    Class to calculate binodal

    Parameters
    ----------
    ref : lnPiMasked
        object to reweight
    build_phases : callable
        function to create Phases object
    build_kws : dict, optional
        optional arguments to build_phases
    """

    collection: lnPiCollection

    def __init__(
        self,
        build_phases: BuildPhasesBase,
        ref: lnPiMasked | None = None,
        build_kws: Mapping[str, Any] | None = None,
    ) -> None:
        self.build_phases = build_phases
        self.ref = ref
        self.build_kws = build_kws or {}

    def objective(self, x: float) -> float | MyNDArray:
        self.collection = self.build_phases(x, ref=self.ref, **self.build_kws)
        out = (
            self.collection.xge.betaOmega()
            .reindex(phase=self.ids)
            .diff("phase")
            .squeeze()
            .to_numpy()
        )

        return array_to_scalar(out)

    def solve(
        self, ids: tuple[int, int], lnz_min: float, lnz_max: float, **kws: Any
    ) -> tuple[lnPiCollection, RootResultTotal]:
        """
        Calculate binodal point where Omega[ID[0]]==Omega[ID[1]]

        Parameters
        ----------
        ids : tuple
            phase index of pair to equate
        lnz_min, lnz_max : float
            lnz_index values bracketing solution
        **kwargs :
            extra arguments to ``brentq``

        Returns
        -------
        binodal : lnPiCollection
            At binodal point.
        stats : object, optional
            Solve stats object from ``brentq`` (optional, returned if full_output is True)

        See Also
        --------
        scipy.optimize.brentq
        """

        from scipy.optimize import brentq

        if len(ids) != 2:
            msg = f"{ids=} must have length 2."
            raise ValueError(msg)

        self.ids = ids

        a, b = min(lnz_min, lnz_max), max(lnz_min, lnz_max)

        xx, r = brentq(self.objective, a, b, full_output=True, **kws)

        return self.collection, _rootresults_to_rootresulttotal(
            r,
            residual=self.objective(xx),  # type: ignore[arg-type]
        )


################################################################################
# Accessor classes/routines
class StabilityBase:
    """
    Base class for stability

    Parameters
    ----------
    collection: lnPiCollection
        Used to bracket the location limit of stability

    """

    _NAME = "base"

    _items: dict[int, lnPiCollection | None]
    _info: dict[int, RootResultTotal]

    def __init__(self, collection: lnPiCollection) -> None:
        self._parent = collection
        self.access_kws: dict[str, Any] = {}
        self._cache: dict[str, Any] = {}

    @property
    def parent(self) -> lnPiCollection:
        """Accessor to parent :class:`~lnpy.lnpiseries.lnPiCollection` object"""
        return self._parent

    def set_access_kws(self, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            self.access_kws[k] = v

    @property
    def items(self) -> dict[int, lnPiCollection | None]:
        """Access to the underlying data"""
        return self._items

    def _get_access(
        self,
        items: Mapping[int, lnPiCollection | None] | None = None,
        concat_kws: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ) -> lnPiCollection:
        items = items or self._items
        concat_kws = concat_kws or {}

        concat_kws = dict(names=[self._NAME], **concat_kws)
        kwargs = dict(self.access_kws, **kwargs)
        # return self._parent.concat_like(items, **concat_kws)
        # s = pd.concat({k:v._series for k,v in items}, **concat_kws)
        # return self._parent.new_like(s)

        return self._parent.concat(items, concat_kws=concat_kws, **kwargs)  # type: ignore[arg-type]

    @cached.prop
    def access(self) -> lnPiCollection:
        """View (:class:`lnpy.lnpiseries.lnPiCollection`) of stability"""
        return self._get_access()

    def __getitem__(self, idx: int) -> lnPiCollection | None:
        return self._items[idx]

    def _get_appender(self, s: lnPiCollection | None = None) -> lnPiCollection:
        if s is None:
            s = self.access
        return s.droplevel(self._NAME)

    @property
    def appender(self) -> lnPiCollection:
        """View of :attr:`access` to be appended to :attr:`parent`"""
        return self._get_appender()

    def append_to_parent(
        self, sort: bool = True, copy_stability: bool = True
    ) -> lnPiCollection:
        """New collection with spinodal data appended to parent."""
        new = self._parent.append(self.appender)
        if sort:
            new = new.sort_index()
        if copy_stability:
            # workaround to copy over spinodal
            new._cache["spinodal"] = self
        return new


# NOTE : single create means this is only created once
# @lnPiCollection.decorate_accessor("spinodal", single_create=False)
class Spinodals(StabilityBase):
    """Methods for calculation locations of spinodal"""

    _NAME = "spinodal"

    _items: dict[int, lnPiCollection | None]
    _info: dict[int, RootResultTotal]

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: Literal[True] = ...,
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> Self: ...

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: Literal[False],
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> tuple[
        dict[int, lnPiCollection | None] | lnPiCollection, dict[int, RootResultTotal]
    ]: ...

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: bool,
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> (
        Self
        | tuple[
            dict[int, lnPiCollection | None] | lnPiCollection,
            dict[int, RootResultTotal],
        ]
    ): ...

    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: bool = True,
        ref: lnPiMasked | None = None,
        build_kws: Mapping[str, Any] | None = None,
        force: bool = False,
        as_dict: bool = True,
        unstack: bool | None = None,
        raise_unconverged: bool = True,
        **kwargs: Any,
    ) -> (
        Self
        | tuple[
            dict[int, lnPiCollection | None] | lnPiCollection,
            dict[int, RootResultTotal],
        ]
    ):
        """
        Parameters
        ----------
        phase_ids : int, or sequence
            Phase ids to calculate spinodal for.  If int, then find spinodals for phases
            range(phase_ids).
        build_phases : callable
            Factory function to build phases.
            This should most likely be an instance of :class:`lnpy.segment.BuildPhasesBase`
        ref : lnPiMasked, optional
        build_kws : dict, optional
            optional arguments to `build_phases`
        inplace : bool, default=True
            if True, add spinodal inplace, otherwise return spinodal object
        force : bool, optional
            if True, force recalculation of spinodal if already set inplace.  Otherwise return already
            calculated values
        as_dict : bool, default=True
            if True, return dict of form {phase_id[0] : phases object, ...}.
        unstack : bool, optional
            if passed, create lnPiCollection objects with this unstack value.  If not passed,
            use unstack parameter from parent object
        raise_unconverged : bool, default=True
            If True, raise error if calculation does not converge.
        **kwargs :
            Extra argument to :meth:`get_spinodal` function

        Returns
        -------
        out : dict or lnPiCollection, optional
            if inplace, return self.
            if not inplace, and as dict, return dict, else return :class:`lnpy.lnpiseries.lnPiCollection` with phase_id in index
        """

        if hasattr(self, "_items") and not force:
            if inplace:
                return self
            if as_dict:
                return self._items, self._info
            return self.access, self._info

        from .segment import BuildPhasesBase

        if not isinstance(build_phases, BuildPhasesBase):
            msg = (
                "`build_phases` should be an instance of `BuildPhasesBase`."
                "Its likely an instance of `PhaseCreator.builphases`."
                "Instead, use an instance of `PhaseCreator.buildphases_mu`."
            )
            raise TypeError(msg)

        if isinstance(phase_ids, int):
            phase_ids = list(range(phase_ids))
        else:
            phase_ids = list(phase_ids)

        out: dict[int, lnPiCollection | None] = {}
        info: dict[int, RootResultTotal] = {}
        for idx in phase_ids:
            s, r = get_spinodal(
                ref=ref,
                collection=self._parent,
                idx=idx,
                build_phases=build_phases,
                build_kws=build_kws,
                **kwargs,
            )
            out[idx] = s
            info[idx] = r

        if unstack is None:
            unstack = self._parent._xarray_unstack
        self.set_access_kws(unstack=unstack)

        converged = all(x["converged"] for x in info.values())
        if raise_unconverged and not converged:
            msg = "Spinodal calculation did not converge"
            raise ValueError(msg)

        if inplace:
            self._items = out
            self._info = info
            # if append:
            #     return self._append_to_parent()
            # else:
            return self

        # total convergence:
        if not as_dict and converged:
            return self._get_access(out), info
        return out, info


# @lnPiCollection.decorate_accessor("binodal", single_create=False)
class Binodals(StabilityBase):
    """Routines to calculate binodal."""

    _NAME = "binodal"
    _items: dict[int, lnPiCollection]  # type: ignore[assignment]

    def get_pair(
        self,
        ids: tuple[int, int],
        solver: _SolveBinodal,
        lnz_min: float | None = None,
        lnz_max: float | None = None,
        spinodals: Spinodals | None = None,
        **kwargs: Any,
    ) -> tuple[lnPiCollection, RootResultTotal]:
        if None in {lnz_min, lnz_max} and spinodals is None:
            spinodals = self._parent.spinodal

        def _get_lnz(idx: int) -> float:
            if spinodals is None:
                raise ValueError
            s = spinodals[idx]
            if s is None:
                msg = f"spinodal with index={idx} is None"
                raise ValueError(msg)
            return s._get_lnz(solver.build_phases.index)

        lnz_min = lnz_min or _get_lnz(ids[0])
        lnz_max = lnz_max or _get_lnz(ids[1])

        return solver.solve(ids=ids, lnz_min=lnz_min, lnz_max=lnz_max, **kwargs)

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: Literal[True] = ...,
        spinodals: Spinodals | None = ...,
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> Self: ...

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: Literal[False],
        spinodals: Spinodals | None = ...,
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> tuple[
        dict[int, lnPiCollection] | lnPiCollection, dict[int, RootResultTotal]
    ]: ...

    @overload
    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: bool,
        spinodals: Spinodals | None = ...,
        ref: lnPiMasked | None = ...,
        build_kws: Mapping[str, Any] | None = ...,
        force: bool = ...,
        as_dict: bool = ...,
        unstack: bool | None = ...,
        raise_unconverged: bool = ...,
        **kwargs: Any,
    ) -> (
        Self
        | tuple[dict[int, lnPiCollection] | lnPiCollection, dict[int, RootResultTotal]]
    ): ...

    def __call__(
        self,
        phase_ids: int | Iterable[int],
        build_phases: BuildPhasesBase,
        *,
        inplace: bool = True,
        spinodals: Spinodals | None = None,
        ref: lnPiMasked | None = None,
        build_kws: Mapping[str, Any] | None = None,
        force: bool = False,
        as_dict: bool = True,
        unstack: bool | None = None,
        raise_unconverged: bool = True,
        **kwargs: Any,
    ) -> (
        Self
        | tuple[dict[int, lnPiCollection] | lnPiCollection, dict[int, RootResultTotal]]
    ):
        """
        Parameters
        ----------
        phase_ids : int, or sequence
            Phase ids to calculate binodal for.  If int, then find binodal for phases
            range(phase_ids).
        build_phases : callable
            Factory function to build phases
        inplace : bool, default=True
            if True, add binodal inplace, otherwise return spinodal object
        spinodals : optional
            if not passes, then use parent.spinodal
            Used for bounding binodal
        ref : lnPiMasked, optional
        build_kws : dict, optional
            optional arguments to `build_phases`
        force : bool, optional
            if True, force recalculation of spinodal if already set inplace.  Otherwise return already
            calculated values
        as_dict : bool, default=True
            if True, return dict of form {phase_id[0] : phases object, ...}
        unstack : bool, optional
            if passed, create lnPiCollection objects with this unstack value.  If not passed,
            use unstack parameter from parent object
        raise_unconverged : bool, default=True
            If True, raise error if calculation does not converge.
        **kwargs
            Extra argument to :meth:`get_spinodal` function

        Returns
        -------
        out : object
            if inplace, return self
            if not inplace, and as dict, return dict, else return lnPiCollection with phase_id in index
        """

        if inplace and hasattr(self, "_items") and not force:
            if inplace:
                return self

            out = self._items if as_dict else self.access
            return out, self._info

        self._solver = _SolveBinodal(
            ref=ref, build_phases=build_phases, build_kws=build_kws
        )

        if isinstance(phase_ids, int):
            phase_ids = list(range(phase_ids))
        else:
            phase_ids = list(phase_ids)

        out = {}
        info = {}
        index = {}
        for idx, ids in enumerate(itertools.combinations(phase_ids, 2)):
            s, r = self.get_pair(
                ids=ids,
                spinodals=spinodals,
                solver=self._solver,
                **kwargs,
            )
            out[idx] = s
            info[idx] = r
            index[idx] = ids

        converged = all(x["converged"] for x in info.values())
        if raise_unconverged and not converged:
            msg = "Binodal calculation did not converge"
            raise ValueError(msg)

        if unstack is None:
            unstack = self._parent._xarray_unstack
        self.set_access_kws(unstack=unstack)

        # either build output or inplace
        if inplace:
            self._items = out
            self._info = info
            self._index = index
            return self

        if not as_dict and converged:
            return self._get_access(out), info
        return out, info


# @lnPiCollection.decorate_accessor("stability_append", single_create=False)
# def _stability_append(self):
#     """
#     Add stability from a collection to this collection

#     Parameters
#     ----------
#     other : optional, optional
#         if passed, copy stability from this collection to self, otherwise
#         use `self`
#     append: bool, default=True
#         if True, append results to new frame
#     sort: bool, default=True
#         if True, sort appended results
#     copy_stability
#     """
#     # Duplicate docstring so show up in docs

#     def func(other=None, append=True, sort=True, copy_stability=True):
#         """
#         Add stability from collection to this collection

#         Parameters
#         ----------
#         other : optional, optional
#             if passed, copy stability from this collection to self, otherwise
#             use `self`
#         append: bool, default=True
#             if True, append results to new frame
#         sort: bool, default=True
#             if True, sort appended results
#         copy_stability
#         """

#         if (not append) and (not copy_stability):
#             raise ValueError("one of append or copy_stability must be True")

#         if other is None:
#             other = self
#         spin = other.spinodal
#         bino = other.binodal
#         if append:
#             new = self.append(spin.appender).append(bino.appender)
#             if sort:
#                 new = new.sort_index()
#         else:
#             new = self.copy()
#         if copy_stability:
#             # TODO(wpk): fix this hack
#             new._cache["spinodal"] = spin
#             new._cache["binodal"] = bino
#             # new.spinodal = spin
#             # new.binodal = bino
#         return new

#     return func



"""
Utility functions (:mod:`~lnpy.utils`)
======================================
"""

from __future__ import annotations

from functools import lru_cache
from itertools import starmap
from typing import TYPE_CHECKING, TypedDict, cast, overload

import numpy as np

from .docstrings import docfiller
from .options import OPTIONS

if TYPE_CHECKING:
    from types import ModuleType
    from typing import Any, Callable, Hashable, Iterable, Mapping, Sequence

    import xarray as xr
    from numpy.typing import ArrayLike, DTypeLike, NDArray
    from scipy.optimize import RootResults

    from lnpy.lnpidata import lnPiMasked

    from ._typing import MaskConvention, MyNDArray, R, T


# --- TQDM setup -----------------------------------------------------------------------
@lru_cache
def _get_tqdm() -> ModuleType | None:
    try:
        import tqdm
    except ImportError:
        return None
    else:
        return tqdm


@lru_cache
def _get_tqdm_default() -> Callable[..., Any]:
    _tqdm = _get_tqdm()
    if _tqdm:
        try:
            from IPython.core.getipython import (  # pyright: ignore[reportMissingImports]
                get_ipython,
            )

            p = get_ipython()  # type: ignore[no-untyped-call, unused-ignore]
            if p is not None and p.has_trait("kernel"):
                from tqdm.notebook import tqdm as tqdm_default

                return tqdm_default
            return cast("Callable[..., Any]", _tqdm.tqdm)
        except ImportError:
            return cast("Callable[..., Any]", _tqdm.tqdm)
    else:

        def wrapper(seq: Iterable[T], *args: Any, **kwargs: Any) -> Iterable[T]:  # noqa: ARG001
            return seq

        return wrapper


def tqdm(seq: Iterable[T], *args: Any, **kwargs: Any) -> Iterable[T]:
    opt = OPTIONS["tqdm_bar"]
    _tqdm = _get_tqdm()
    if _tqdm:
        if opt == "text":
            func = _tqdm.tqdm
        elif opt == "notebook":
            func = _tqdm.tqdm_notebook
        else:
            func = _get_tqdm_default()
    else:
        func = _get_tqdm_default()
    return cast("Iterable[T]", func(seq, *args, **kwargs))


def get_tqdm(
    seq: Iterable[T], len_min: str | int, leave: bool | None = None, **kwargs: Any
) -> Iterable[T]:
    n = kwargs.get("total", None)
    _tqdm = _get_tqdm()

    if isinstance(len_min, str):
        len_min = OPTIONS[len_min]  # type: ignore[literal-required]

        if not isinstance(len_min, int):
            msg = f"{type(len_min)=} must be an int or a string in options."
            raise TypeError(msg)

    if n is None:
        seq = tuple(seq)
        n = len(seq)

    if _tqdm and OPTIONS["tqdm_use"] and n >= len_min:
        if leave is None:
            leave = OPTIONS["tqdm_leave"]
        seq = tqdm(seq, leave=leave, **kwargs)
    return seq


def get_tqdm_calc(
    seq: Iterable[T], leave: bool | None = None, **kwargs: Any
) -> Iterable[T]:
    return get_tqdm(seq, len_min="tqdm_len_calc", leave=leave, **kwargs)


def get_tqdm_build(
    seq: Iterable[T], leave: bool | None = None, **kwargs: Any
) -> Iterable[T]:
    return get_tqdm(seq, len_min="tqdm_len_build", leave=leave, **kwargs)


# get_tqdm_calc = partial(get_tqdm, len_min="tqdm_len_calc")
# get_tqdm_build = partial(get_tqdm, len_min="tqdm_len_build")

# def get_tqdm_calc(seq, len_min=None, leave=None, **kwargs):
#     if len_min is None:
#         len_min = OPTIONS['tqdm_len_calc']
#     return _get_tqdm(seq, len_min=len_min, leave=leave, **kwargs)


# def get_tqdm_build(seq, len_min=None, leave=None, **kwargs):
#     if len_min is None:
#         len_min = OPTIONS['tqdm_len_build']
#     return _get_tqdm(seq, len_min=len_min, leave=leave, **kwargs)


# --------------------------------------------------
# JOBLIB stuff
@lru_cache
def _get_joblib() -> Any:
    try:
        import joblib
    except ImportError:
        joblib = None
    return joblib


def _use_joblib(
    items: Sequence[Any],
    len_key: str,
    use_joblib: bool = True,
    total: int | None = None,
) -> bool:
    joblib = _get_joblib()

    if use_joblib and joblib and OPTIONS["joblib_use"]:
        if total is None:
            total = len(items)

        return total >= OPTIONS[len_key]  # type: ignore[no-any-return,literal-required]
    return False


def _parallel(seq: Iterable[Any]) -> list[Any]:
    joblib = _get_joblib()

    return joblib.Parallel(  # type: ignore[no-any-return]
        n_jobs=OPTIONS["joblib_n_jobs"],
        backend=OPTIONS["joblib_backend"],
        **OPTIONS["joblib_kws"],
    )(seq)


def parallel_map_build(
    func: Callable[..., R], items: Iterable[Any], *args: Any, **kwargs: Any
) -> list[R]:
    joblib = _get_joblib()

    items = tuple(items)

    if _use_joblib(items, "joblib_len_build"):
        return _parallel(joblib.delayed(func)(x, *args, **kwargs) for x in items)
    return [func(x, *args, **kwargs) for x in items]


def _func_call(x: Callable[..., R], *args: Any, **kwargs: Any) -> R:
    return x(*args, **kwargs)


def parallel_map_call(
    items: Sequence[Callable[..., Any]],
    use_joblib: bool,  # noqa: ARG001
    *args: Any,
    **kwargs: Any,
) -> list[Any]:
    joblib = _get_joblib()
    if _use_joblib(items, "joblib_len_calc"):
        return _parallel(joblib.delayed(_func_call)(x, *args, **kwargs) for x in items)
    return [x(*args, **kwargs) for x in items]


def parallel_map_attr(attr: str, use_joblib: bool, items: Sequence[Any]) -> list[Any]:  # noqa: ARG001
    from operator import attrgetter

    joblib = _get_joblib()
    func = attrgetter(attr)

    if _use_joblib(items, "joblib_len_calc"):
        return _parallel(joblib.delayed(func)(x) for x in items)
    return [func(x) for x in items]


def parallel_map_func_starargs(
    func: Callable[..., R],
    use_joblib: bool,  # noqa: ARG001
    items: Iterable[Any],
    total: int | None = None,
) -> list[R]:
    joblib = _get_joblib()

    items = tuple(items)

    if _use_joblib(items, "joblib_len_calc", total=total):
        return _parallel(starmap(joblib.delayed(func), items))
    return list(starmap(func, items))


# ----------------------------------------
# pandas stuff
def allbut(levels: Iterable[str], *names: str) -> list[str]:
    name_set = set(names)
    return [item for item in levels if item not in name_set]


# ----------------------------------------
# xarray utils
def dim_to_suffix_dataarray(
    da: xr.DataArray, dim: Hashable, join: str = "_"
) -> xr.Dataset:
    if dim in da.dims:
        return da.assign_coords(  # type: ignore[misc]
            **{dim: lambda x: [f"{x.name}{join}{c}" for c in x[dim].to_numpy()]}  # type: ignore[arg-type]
        ).to_dataset(dim=dim)
    return da.to_dataset()


def dim_to_suffix_dataset(
    table: xr.Dataset, dim: Hashable, join: str = "_"
) -> xr.Dataset:
    out = table
    for k in out:
        if dim in out[k].dims:
            out = out.drop_vars(k).update(  # type: ignore[arg-type,unused-ignore]
                table[k].pipe(dim_to_suffix_dataarray, dim, join)
            )
    return out


@overload
def dim_to_suffix(
    ds: xr.DataArray, dim: Hashable = ..., join: str = ...
) -> xr.DataArray: ...


@overload
def dim_to_suffix(
    ds: xr.Dataset, dim: Hashable = ..., join: str = ...
) -> xr.Dataset: ...


def dim_to_suffix(
    ds: xr.DataArray | xr.Dataset, dim: Hashable = "component", join: str = "_"
) -> xr.DataArray | xr.Dataset:
    from xarray import DataArray, Dataset

    if isinstance(ds, DataArray):
        return dim_to_suffix_dataarray(ds, dim=dim, join=join)
    if isinstance(ds, Dataset):
        return dim_to_suffix_dataset(ds, dim=dim, join=join)
    msg = "`ds` must be `DataArray` or `Dataset`"
    raise ValueError(msg)


def _convention_to_bool(convention: MaskConvention) -> bool:
    if convention == "image":
        convention = True
    elif convention == "masked":
        convention = False
    elif not isinstance(convention, bool):
        msg = f"Bad value {convention} sent to _convention_to_bool"
        raise ValueError(msg)
    return convention


@overload
def mask_change_convention(
    mask: None,
    convention_in: MaskConvention = ...,
    convention_out: MaskConvention = ...,
) -> None: ...


@overload
def mask_change_convention(
    mask: MyNDArray,
    convention_in: MaskConvention = ...,
    convention_out: MaskConvention = ...,
) -> MyNDArray: ...


@docfiller.decorate
def mask_change_convention(
    mask: MyNDArray | None,
    convention_in: MaskConvention = "image",
    convention_out: MaskConvention = "masked",
) -> MyNDArray | None:
    """
    Convert an array from one 'mask' convention to another.

    Parameters
    ----------
    {mask_general}
    convention_in, convention_out : string or bool
        Convention for input and output.
        Convention for mask.  Allowable values are:

        * ``'image' ``or ``True`` : `True` values included, `False` values excluded.
          This is the normal convention in :mod:`scipy.ndimage`.
        * ``'masked'`` or ``False``:  `False` values are included, `True` values are excluded.
          This is the convention in :mod:`numpy.ma`

    Returns
    -------
    ndarray
        New 'mask' array with specified convention.
    """

    if mask is None:
        return mask

    convention_in = _convention_to_bool(convention_in)
    convention_out = _convention_to_bool(convention_out)

    if convention_in != convention_out:
        mask = ~mask
    return mask


# m0: list[MyNDArray]
# reveal_type(masks_change_convention(m0))

# m1: list[None]
# reveal_type(masks_change_convention(m1))

# m2: list[MyNDArray | None]
# reveal_type(masks_change_convention(m2))


@overload
def masks_change_convention(
    masks: Sequence[MyNDArray],
    convention_in: MaskConvention = ...,
    convention_out: MaskConvention = ...,
) -> Sequence[MyNDArray]: ...


@overload
def masks_change_convention(
    masks: Sequence[None],
    convention_in: MaskConvention = ...,
    convention_out: MaskConvention = ...,
) -> Sequence[None]: ...


@overload
def masks_change_convention(
    masks: Sequence[MyNDArray | None],
    convention_in: MaskConvention = ...,
    convention_out: MaskConvention = ...,
) -> Sequence[MyNDArray | None]: ...


def masks_change_convention(
    masks: Sequence[MyNDArray] | Sequence[None] | Sequence[MyNDArray | None],
    convention_in: MaskConvention = "image",
    convention_out: MaskConvention = "masked",
) -> Sequence[MyNDArray] | Sequence[None] | Sequence[MyNDArray | None]:
    """
    Perform convention change of sequence of masks

    Parameters
    ----------
    masks : sequence of array-like
        masks[i] is the 'ith' mask
    convention_in, convention_out : string or bool or None
        Convention for input and output.
        Convention for mask.  Allowable values are:

        * 'image' or True : `True` values included, `False` values excluded.
          This is the normal convention in :mod:`scipy.ndimage`.
        * 'masked' or False:  `False` values are included, `True` values are excluded.
          This is the convention in :mod:`numpy.ma`

        If `None`, then pass return input `mask`

    Returns
    -------
    new_masks : list of array
        New 'masks' array with specified convention.
    """

    convention_in = _convention_to_bool(convention_in)
    convention_out = _convention_to_bool(convention_out)

    if convention_in != convention_out:
        masks = [m if m is None else ~m for m in masks]
    return masks


##################################################
# labels/masks utilities
##################################################
@docfiller.decorate
def labels_to_masks(
    labels: MyNDArray,
    features: Sequence[int] | MyNDArray | None = None,
    include_boundary: bool = False,
    convention: MaskConvention = "image",
    check_features: bool = True,
    **kwargs: Any,
) -> tuple[list[MyNDArray], Sequence[int]]:
    """
    Convert labels array to list of masks

    Parameters
    ----------
    labels : ndarray of int
        Each unique value `i` in `labels` indicates a mask.
        That is ``labels == i``.
    {features}
    {include_boundary}
    {mask_convention}
    {check_features}

    **kwargs
        arguments to find_boundary if include_boundary is True
        default to mode='outer', connectivity=labels.ndim

    Returns
    -------
    output : list of array of bool
        list of mask arrays, each with same shape as ``labels``, with
        mask convention `convention`.
    features : list
        features

    See Also
    --------
    masks_to_labels
    :func:`skimage.segmentation.find_boundaries`

    """
    from skimage import segmentation

    if include_boundary:
        kwargs = dict({"mode": "outer", "connectivity": labels.ndim}, **kwargs)
    if features is None:
        features = [i for i in np.unique(labels) if i > 0]

    elif check_features:
        vals = np.unique(labels)
        if not np.all([x in vals for x in features]):
            raise ValueError

    convention = _convention_to_bool(convention)

    output: list[MyNDArray] = []
    for i in features:
        m: NDArray[np.bool_] = labels == i
        if include_boundary:
            # fmt: off
            b = cast(
                "NDArray[np.bool_]",
                segmentation.find_boundaries(m.astype(int), **kwargs),  # pyright: ignore[reportUnknownMemberType]
            )
            # fmt: on
            m = m | b
        if not convention:
            m = ~m
        output.append(m)
    return output, features  # type: ignore[return-value]


@docfiller.decorate
def masks_to_labels(
    masks: Sequence[MyNDArray],
    features: Sequence[int] | MyNDArray | None = None,
    convention: MaskConvention = "image",
    dtype: DTypeLike = np.int_,
) -> MyNDArray:
    """
    Convert list of masks to labels

    Parameters
    ----------
    masks : list of array-like of bool
        list of mask arrays.
    {features}
    {mask_convention}

    Returns
    -------
    ndarray
        Label array.


    See Also
    --------
    labels_to_masks
    """

    if features is None:
        features = range(1, len(masks) + 1)
    elif len(features) != len(masks):
        raise ValueError

    labels = np.full(masks[0].shape, fill_value=0, dtype=dtype)

    masks = masks_change_convention(masks, convention, True)

    for i, m in zip(features, masks):
        labels[m] = i
    return labels


def ffill(arr: MyNDArray, axis: int = -1, limit: int | None = None) -> MyNDArray:
    import bottleneck

    _limit = limit if limit is not None else arr.shape[axis]
    return bottleneck.push(arr, n=_limit, axis=axis)  # type: ignore[no-any-return]


def bfill(arr: MyNDArray, axis: int = -1, limit: int | None = None) -> MyNDArray:
    """Inverse of ffill"""
    import bottleneck

    # work around for bottleneck 178
    _limit = limit if limit is not None else arr.shape[axis]

    arr = np.flip(arr, axis=axis)
    # fill
    arr = bottleneck.push(arr, axis=axis, n=_limit)
    # reverse back to original
    return np.flip(arr, axis=axis)


##################################################
# calculations
##################################################


def get_lnz_iter(lnz: Iterable[float | None], x: ArrayLike) -> MyNDArray:
    """
    Create a lnz_iter object for varying a single lnz

    Parameters
    ----------
    lnz : list
        list with one element equal to None.  This is the component which will be varied
        For example, lnz=[lnz0,None,lnz2] implies use values of lnz0,lnz2 for components 0 and 2, and
        vary component 1

    x : array
        values to insert for variable component

    Returns
    -------
    ndarray
        Shape ``(len(x),len(lnz))``.
        array with rows [lnz0,lnz1,lnz2]
    """
    x = np.asarray(x)
    z = np.zeros_like(x)

    return np.array([x if m is None else z + m for m in lnz]).T


##################################################
# utilities
##################################################


# def sort_lnPis(values: Sequence[lnPiMasked], comp=0) -> list[lnPiMasked]:
#     """
#     Sort list of lnPi  that component `comp` mole fraction increases

#     Parameters
#     ----------
#     values : list of lnPiMasked

#     comp : int, default=0
#      component to sort along

#     Returns
#     -------
#     output : list of lnPiMasked
#         Objects in sorted order.
#     """

#     molfrac_comp = np.array([x.molfrac[comp] for x in values])

#     order = np.argsort(molfrac_comp)

#     output = [values[i] for i in order]

#     return output


def distance_matrix(
    mask: ArrayLike, convention: MaskConvention = "image"
) -> NDArray[np.float64]:
    """
    Create matrix of distances from elements of mask
    to nearest background point

    Parameters
    ----------
    mask : array-like
        image mask
    convention : str or bool, default='image'
        mask convention

    Returns
    -------
    distance : ndarray
        Same shape as mask.
        Distance from possible feature elements to background


    See Also
    --------
    ~scipy.ndimage.distance_transform_edt
    """
    from scipy.ndimage import distance_transform_edt

    mask = np.asarray(mask, dtype=bool)
    mask = mask_change_convention(mask, convention_in=convention, convention_out=True)

    # pad mask
    # add padding to end of matrix in each dimension
    ndim = mask.ndim
    pad_width = ((0, 1),) * ndim
    mask = np.pad(mask, pad_width=pad_width, mode="constant", constant_values=False)

    # distance filter
    dist = distance_transform_edt(mask)

    # remove padding
    s = (slice(None, -1),) * ndim
    return dist[s]  # type: ignore[no-any-return]


def lnpimasked_to_dataset(
    data: lnPiMasked, keys: Sequence[str] = ("lnpi", "PE")
) -> xr.Dataset:
    """
    Convert a :class:`~lnpy.lnpidata.lnPiMasked` object into as :class:`~xarray.Dataset`.

    Parameters
    ----------
    data : lnPiMasked

    Returns
    -------
    output : Dataset
    """

    return data.xce.table(keys=keys, default_keys=None)


def dataset_to_lnpimasked(
    ds: xr.Dataset,
    lnpi_name: str = "lnpi",
    pe_name: str = "PE",
    extra_kws: Mapping[str, Any] | None = None,
    **kwargs: Any,
) -> lnPiMasked:
    """
    Convert a :class:`~xarray.Dataset` to a :class:`~lnpy.lnpidata.lnPiMasked` object.

    Parameters
    ----------
    ds : :class:`~xarray.Dataset`
    lnpi_name, pe_name : str
        Names of 'lnPi' and 'PE' parameters
    extra_kws : mapping, optional
        parameter `extra_kws`.  Note that if `pe_name` found in `ds`, then will add it to `extra_kws`.

    Returns
    -------
    lnpi : lnPiMasked
    """

    from .lnpidata import lnPiMasked

    data = ds[lnpi_name]

    if extra_kws is None:
        extra_kws = {}

    if pe_name in ds:
        extra_kws[pe_name] = ds[pe_name].to_numpy()  # type: ignore[index]

    return lnPiMasked.from_dataarray(da=data, extra_kws=extra_kws, **kwargs)


# --- Root results ---------------------------------------------------------------------


class _RootResultDictReq(TypedDict, total=True):
    root: float | MyNDArray | None
    iterations: int
    function_calls: int
    converged: bool
    flag: str


class RootResultDict(_RootResultDictReq, total=False):
    """Interface to :class:`scipy.optimize.RootResults`."""

    residual: float | MyNDArray


def rootresults_to_rootresultdict(
    r: RootResults, residual: float | MyNDArray | None
) -> RootResultDict:
    """Convert :class:`scipy.optimize.RootResults` to typed dictionary"""

    out = RootResultDict(
        root=r.root,
        iterations=r.iterations,
        function_calls=r.function_calls,
        converged=r.converged,
        flag=r.flag,  # pyright: ignore[reportArgumentType]
    )

    if residual is not None:
        out["residual"] = residual
    return out


def array_to_scalar(x: float | MyNDArray) -> float:
    """
    Convert array to scalar.

    If `x` is an ndarray, convert to float of 0-d,
    or extract first element (of flattened array )
    if N-d.
    """
    if isinstance(x, np.ndarray):
        return x.flat[0]  # type: ignore[no-any-return]
    return x



"""Set of helper utilities to work with single component system"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np

from .lnpiseries import lnPiCollection
from .utils import RootResultDict, array_to_scalar, rootresults_to_rootresultdict

if TYPE_CHECKING:
    from typing import Any, Mapping, Sequence

    import xarray as xr
    from numpy.typing import ArrayLike

    from ._typing import MyNDArray
    from .lnpidata import lnPiMasked
    from .segment import BuildPhasesBase


def tag_phases_singlecomp(x: Sequence[lnPiMasked]) -> Sequence[int] | MyNDArray:
    """
    Function to tag phases with a unique id.

    This is for analyzing a single component system.
    If multiple phases passed (length of input > 1), then
    sort by ``argmax = x.local_argmax()``, i.e., the location of the maxima.
    Otherwise, assign `phase_id = 0` if ``argmax < Nmax // 2`` where ``Nmax`` is
    the maximum number of particles in lnPi.

    Parameters
    ----------
    x : sequence of lnPiMasked
        lnPi objects to be tagged.

    Returns
    -------
    phase_id : sequence of int
        Phase code for each element in ``x``.
    """
    if len(x) > 2:
        msg = "bad tag function"
        raise ValueError(msg)

    argmax0 = np.array([xx.local_argmax()[0] for xx in x])
    if len(x) == 2:
        return np.argsort(argmax0)

    return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)


def get_lnz_min(
    target: float,
    collection: lnPiCollection,
    build_phases: BuildPhasesBase,
    phase_id: int | None = 0,
    component: int | None = None,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
    ntry: int = 20,
    solve_kws: Mapping[str, Any] | None = None,
) -> tuple[lnPiCollection, RootResultDict]:
    """
    Solve for target density

    Parameters
    ----------
    target : float
        target density of specified component
    collection : CollectionPhases
        initial guess to work from.  This is assumed to be sorted in ascending `lnz` order
    build_phases : callable
        function to build new phases objects
    phase_id : int, default=0
        phase id tag to consider.  If None, consider min over 'phases'
    component : int, optional
        component to consider.
        If component is None, use build_phases.index.
        If component=='None', then consider total density
    dlnz : float, default=0.5
        how to change the chemical potential if `collection` doesn't already
        bracket solution
    ref : lnPiMasked, optional
        optional lnPiMasked object to pass to build_phases
    build_kws : dictionary, optional
        optional arguments to `build_phases`
    ntry : int, default=20
        number of times to attempted finding left/right bracketing objects
    solve_kws : dictionary, optional
        optional arguments to `scipy.optimize.brentq`

    Returns
    -------
    phases : Phases object as solution lnz
        solution_parameters : optional
    info : :class:`RootResultDict`
    """
    from scipy.optimize import brentq

    if phase_id is not None and phase_id not in collection.index.get_level_values(
        "phase"
    ):
        msg = f"no phase {phase_id}"
        raise ValueError(msg)

    lnz_idx = build_phases.index
    if isinstance(component, str) and component.lower() == "none":

        def getter_comp(x: lnPiCollection) -> xr.DataArray:
            return x.xge.dens_tot

    else:
        component = component or lnz_idx

        def getter_comp(x: lnPiCollection) -> xr.DataArray:
            return x.xge.dens.sel(component=component)

    if phase_id is None:

        def getter(x: lnPiCollection) -> xr.DataArray:
            return getter_comp(x).min("phase")

    else:

        def getter(x: lnPiCollection) -> xr.DataArray:
            return getter_comp(x).sel(phase=phase_id)

    s = getter(collection).to_series().dropna()
    # builder
    if build_kws is None:
        build_kws = {}

    # left bracket
    left = None
    ss = s[s < target]
    if len(ss) > 0:
        left = collection.mloc[ss.index[[-1]]]
    else:
        # TODO(wpk): Need to make collection work better.  Might need to just have a single class for everything...
        # reveal_type(collection.mloc[s.index[[0]]])
        new_lnz = collection.mloc[s.index[[0]]]._get_lnz(lnz_idx)
        dlnz_left = dlnz
        for _i in range(ntry):
            new_lnz -= dlnz_left
            p = build_phases(new_lnz, ref=ref, **build_kws)
            if phase_id in p._get_level("phase") and getter(p).to_numpy() < target:
                left = p
                break
            dlnz_left *= dfac

    if left is None:
        msg = "could not find left bounds"
        raise RuntimeError(msg)

    # right bracket
    right = None
    ss = s[s > target]
    if len(ss) > 0:
        right = collection.mloc[ss.index[[0]]]
    else:
        new_lnz = collection.mloc[s.index[[-1]]]._get_lnz(lnz_idx)
        dlnz_right = dlnz

        for _i in range(ntry):
            new_lnz += dlnz_right
            p = build_phases(new_lnz, ref=ref, **build_kws)

            if phase_id not in p._get_level("phase"):
                # went to far
                new_lnz -= dlnz_right
                # reset to half dlnz
                dlnz_right = dlnz_right * 0.5
            elif getter(p).to_numpy() > target:
                right = p
                break
            else:
                dlnz_right *= dfac

    if right is None:
        msg = "could not find right bounds"
        raise RuntimeError(msg)

    def f(x: float) -> float:
        lnz_new = x
        p = build_phases(lnz_new, ref=ref, **build_kws)
        f.lnpi = p  # type: ignore[attr-defined]
        return array_to_scalar(getter(p).values) - target

    a, b = sorted([x._get_lnz(lnz_idx) for x in [left, right]])

    xx, r = brentq(f, a, b, full_output=True, **(solve_kws or {}))

    return f.lnpi, rootresults_to_rootresultdict(r, residual=f(xx))  # type: ignore[attr-defined]


def get_lnz_max(
    edge_distance_min: int,
    build_phases: BuildPhasesBase,
    ref: lnPiMasked | None = None,
    lnz_start: float | None = None,
    collection: lnPiCollection | None = None,
    dlnz: float = 0.5,
    dfac: float = 1.0,
    threshold_abs: float = 1e-4,
    ntry: int = 50,
    build_kws: Mapping[str, Any] | None = None,
) -> tuple[lnPiCollection, dict[str, int | float]]:
    """Find max lnz by bisection"""

    build_kws = build_kws or {}
    ref = ref or build_phases.phase_creator.ref

    if ref is None:
        msg = "must specify `ref` or build_phases must have access to reference lnPiMasked object"
        raise ValueError(msg)

    lnz_idx = build_phases.index
    lnz_start = cast(float, lnz_start or ref.lnz[lnz_idx])

    # need left/right bounds
    # left is greatest lnz point with edge_distance > edge_distance_min
    # right is least lnz point with edge_distance < edge_distance_min
    left: lnPiCollection | None = None
    right: lnPiCollection | None = None
    lnz_left: float | None = None
    lnz_right: float | None = None
    n_left: int | None = None
    n_right: int | None = None

    def getter(p: lnPiCollection) -> xr.DataArray:
        v = p.xge.edge_distance(ref)
        if not p._xarray_unstack:
            v = v.unstack(p._concat_dim)  # noqa: PD010
        return v.min("phase")

    # if have collection, try to use it
    if collection is not None:
        # see if bound contained in collection

        s = getter(collection).to_series()

        # left
        ss = s[s > edge_distance_min]
        if len(ss) > 0:
            left = collection.mloc[ss.index[[-1]]]
        else:
            lnz_left = collection.zloc[[0]]._get_lnz(lnz_idx)

        # right
        ss = s[s < edge_distance_min]
        if len(ss) > 0:
            right = collection.mloc[ss.index[[0]]]
        else:
            lnz_right = collection.zloc[[-1]]._get_lnz(lnz_idx)

    # if left not set, try to find it
    if left is None:
        if lnz_left is None:
            lnz_left = lnz_start
        dlnz_loc = dlnz
        for i in range(ntry):
            lnz_left -= dlnz_loc
            p = build_phases(lnz_left, ref=ref, **build_kws)
            if getter(p).to_numpy() >= edge_distance_min:
                left = p
                n_left = i
                break
            # grow dlnz
            dlnz_loc *= dfac

    if left is None:
        msg = "could not find left bound"
        raise RuntimeError(msg)

    # if right not set, try to find it
    if right is None:
        if lnz_right is None:
            lnz_right = lnz_start

        dlnz_loc = dlnz
        for i in range(ntry):
            lnz_right += dlnz_loc
            p = build_phases(lnz_right, ref=ref, **build_kws)

            if getter(p).to_numpy() < edge_distance_min:
                right = p
                n_right = i
                break
            # grow dlnz
            dlnz_loc *= dfac

    if right is None:
        msg = "could not find right bound"
        raise RuntimeError(msg)

    # not do bisection
    bracket = [left, right]
    values = [getter(x).to_numpy() for x in bracket]

    for i in range(ntry):  # noqa: B007
        lnz = [x._get_lnz(lnz_idx) for x in bracket]
        delta = np.abs(lnz[1] - lnz[0])
        if delta < threshold_abs:
            break
        lnz_mid = 0.5 * (lnz[0] + lnz[1])
        mid = build_phases(lnz_mid, ref=ref, **build_kws)
        y_mid = getter(mid).to_numpy()

        index = 0 if y_mid >= edge_distance_min else 1

        bracket[index] = mid
        values[index] = y_mid

    left, right = bracket

    info = {"ntry": i, "ntry_left": n_left, "ntry_right": n_right, "precision": delta}
    return left, info


def build_grid(
    x: ArrayLike | MyNDArray | None = None,
    dx: float | None = None,
    x_range: Sequence[float] | MyNDArray | None = None,
    x0: float | None = None,
    offsets: Sequence[float] | MyNDArray | None = None,
    even_grid: bool = False,
    digits: int | None = None,
    unique: bool = True,
    outlier: bool = False,
) -> MyNDArray:
    """
    Build a grid of values


    Parameters
    ----------
    x : array, optional
        if passes, use this as the base array
    dx : float, optional
        delta x. See below for uses
    x_range : array of length 2, optional
        if x not specified, make an array with
        `x=np.arange(x_range[0], x_range[1] + dx/2, dx)`
    x0 : float, optional
        see offsets
    offsets : array of length 2, optional
        if x or x_range not specified, set x_range = x0 + offsets
    even_grid : bool, default=True
        if True, make sure all values are multiples of dx
    digits : int, optional
        if passes, round output to this many digits
    outlier : bool, default=False
        if outlier is True, allow values outsize x

    Returns
    -------
    out : array of values
    """

    if x is None:
        if dx is None:
            msg = "Must specify dx"
            raise ValueError(msg)

        if x_range is not None:
            x_range = np.asarray(x_range)
        elif offsets is not None and x0 is not None:
            x_range = x0 + np.asarray(offsets)
        else:
            msg = "must specify x_range or x0 and offsets"
            raise ValueError(msg)

        if len(x_range) != 2:
            raise ValueError

        # for consistency with previous code.
        if even_grid:
            new_range = np.round(x_range / dx) * dx
            if new_range[0] < x_range[0]:
                new_range[0] += dx
            if new_range[1] > x_range[1]:
                new_range[1] -= dx
            x_range = new_range
        x = np.arange(x_range[0], x_range[1] + dx * 0.5, dx)  # type: ignore[index]
    else:
        x = np.asarray(x)

    if not outlier:
        lb, ub = x.min(), x.max()

    if even_grid:
        if dx is None:
            raise TypeError
        x = np.round(x / dx) * dx

    if digits is not None:
        x = np.round(x, digits)  # type: ignore[arg-type]

    if unique:
        x = np.unique(x)  # type: ignore[arg-type]

    if not outlier:
        x = x[(lb <= x) & (x <= ub)]  # type: ignore[index]

    return x  # type: ignore[return-value]


def limited_collection(
    build_phases: BuildPhasesBase,
    dlnz: float,
    lnz_range: Sequence[float] | MyNDArray | None = None,
    offsets: list[int] | None = None,
    digits: int | None = None,
    even_grid: bool = True,
    course_step: int | None = None,
    edge_distance_min: int | None = None,
    dens_min: float | None = None,
    lnz_min_kws: Mapping[str, Any] | None = None,
    lnz_max_kws: Mapping[str, Any] | None = None,
    ref: lnPiMasked | None = None,
    build_kws: Mapping[str, Any] | None = None,
    # build_stability_kws: Mapping[str, Any] | None = None,
    nmax: int | None = None,
    xarray_output: bool = True,
    collection_kws: Mapping[str, Any] | None = None,
    limit_course: bool = False,
) -> tuple[lnPiCollection, lnPiCollection]:
    """
    Build a lnPiCollection over a range of lnz values

    Parameters
    ----------
    build_phases :
    """

    if lnz_range is None:
        ref = ref or build_phases.phase_creator.ref
        if ref is None:
            msg = "Must pass in ref or build_phases must have access to reference lnPiMasked object"
            raise ValueError(msg)
        x0 = ref.lnz[build_phases.index]
    else:
        x0 = None

    # TODO: update tests to get rid of outlier=True  # noqa: TD002
    lnzs = build_grid(
        dx=dlnz,
        x_range=lnz_range,
        x0=x0,
        offsets=offsets,
        even_grid=even_grid,
        digits=digits,
        outlier=True,
    )

    if course_step is None:
        course_step = max(len(lnzs) // 40, 1)
    if collection_kws is None:
        collection_kws = {}

    # limit lnz
    c_course = None
    lnz_min, lnz_max = lnzs[0], lnzs[-1]
    if edge_distance_min is not None or dens_min is not None:
        c_course = lnPiCollection.from_builder(
            lnzs[::course_step],
            build_phases=build_phases,
            build_kws=build_kws,
            nmax=nmax,
            xarray_output=xarray_output,
            **collection_kws,
        )

        if dens_min is not None:
            try:
                if lnz_min_kws is None:
                    lnz_min_kws = {}
                p_min, o = get_lnz_min(
                    dens_min, c_course, build_phases, build_kws=build_kws, **lnz_min_kws
                )
                if o["converged"]:
                    lnz_min = p_min.iloc[0].lnz[build_phases.index] - dlnz
            except Exception:  # noqa: BLE001, S110
                pass

        if edge_distance_min is not None:
            try:
                if lnz_max_kws is None:
                    lnz_max_kws = {}
                p_max, _ = get_lnz_max(
                    edge_distance_min,
                    build_phases,
                    build_kws=build_kws,
                    collection=c_course,
                    **lnz_max_kws,
                )
                lnz_max = p_max.iloc[0].lnz[build_phases.index]
            except Exception:  # noqa: BLE001,S110
                pass

    lnzs = lnzs[(lnzs >= lnz_min) & (lnzs <= lnz_max)]
    c = lnPiCollection.from_builder(
        lnzs,
        build_phases=build_phases,
        build_kws=build_kws,
        nmax=nmax,
        xarray_output=xarray_output,
        **collection_kws,
    )

    if c_course is None:
        c_course = c
    elif limit_course:
        slnz = f"lnz_{build_phases.index}"
        q = f"{slnz} >= {lnz_min} and {slnz} <= {lnz_max}"
        c_course = c_course.query(q)

    return c_course, c
